import os
import json
import torch
import random

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from abc import abstractmethod
from abc import ABC
from PIL import Image
from enum import Enum
from tqdm import tqdm
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import TypeVar
from typing import Callable
from typing import Optional
from typing import NamedTuple
from typing import ContextManager
from cftool.cv import read_image
from cftool.cv import save_images
from cftool.cv import restrict_wh
from cftool.cv import get_suitable_size
from cftool.cv import ReadImageResponse
from cftool.misc import safe_execute
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.types import arr_type
from cftool.types import tensor_dict_type

from ..zoo import ldm_sd
from ..zoo import ldm_sd_v2
from ..zoo import ldm_sd_v2_base
from ..zoo import ldm_sd_tag
from ..zoo import ldm_sd_inpainting
from ..zoo import ldm_sr
from ..zoo import ldm_semantic
from ..zoo import ldm_celeba_hq
from ..zoo import ldm_inpainting
from ..zoo import DLZoo
from ..utils import APIMixin
from ..utils import WeightsPool
from .third_party import MiDaSAPI
from .third_party import MLSDDetector
from .third_party import OpenposeDetector
from ...data import ArrayData
from ...schema import DataConfig
from ...constants import INPUT_KEY
from ...constants import PREDICTIONS_KEY
from ...parameters import OPT
from ...data.utils import predict_array_data
from ...data.utils import TensorBatcher
from ...misc.toolkit import slerp
from ...misc.toolkit import new_seed
from ...misc.toolkit import download_model
from ...misc.toolkit import download_static
from ...misc.toolkit import seed_everything
from ...misc.toolkit import eval_context
from ...modules.blocks import Conv2d
from ...models.cv.diffusion import LDM
from ...models.cv.diffusion import DDPM
from ...models.cv.diffusion import ISampler
from ...models.cv.diffusion import ControlNet
from ...models.cv.diffusion import StableDiffusion
from ...models.cv.ae.common import IAutoEncoder
from ...models.cv.diffusion.utils import get_timesteps
from ...models.cv.diffusion.utils import CONCAT_KEY
from ...models.cv.diffusion.utils import CONCAT_TYPE
from ...models.cv.diffusion.utils import HYBRID_TYPE
from ...models.cv.diffusion.utils import CROSS_ATTN_KEY
from ...models.cv.diffusion.utils import CONTROL_HINT_KEY
from ...models.cv.diffusion.utils import CONTROL_HINT_START_KEY
from ...models.cv.diffusion.cond_models import CLIPTextConditionModel
from ...models.cv.diffusion.samplers.ddim import DDIMMixin
from ...models.cv.diffusion.samplers.solver import DPMSolver
from ...models.cv.diffusion.samplers.k_samplers import KSamplerMixin

try:
    import cv2
except:
    cv2 = None


class switch_sampler_context:
    def __init__(self, api: "DiffusionAPI", sampler: Optional[str]):
        self.api = api
        self.m_sampler = api.sampler.__identifier__
        self.target_sampler = sampler
        self.switched = False

    def __enter__(self) -> None:
        if self.target_sampler is None:
            return
        if self.m_sampler != self.target_sampler:
            self.switched = True
            self.api.switch_sampler(self.target_sampler)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.switched:
            self.api.switch_sampler(self.m_sampler)


class SizeInfo(NamedTuple):
    factor: int
    opt_size: int


class MaskedCond(NamedTuple):
    image_res: ReadImageResponse
    mask_res: ReadImageResponse
    mask: np.ndarray
    remained_image: np.ndarray
    remained_mask: np.ndarray
    mask_cond: Tensor
    remained_image_cond: Tensor


T = TypeVar("T", bound="DiffusionAPI")
TAnnotator = TypeVar("TAnnotator", bound="Annotator")


class SDVersions(str, Enum):
    v1_5_BC = ""
    v1_5 = "v1.5"
    ANIME = "anime"
    ANIME_ANYTHING = "anime_anything"
    ANIME_HYBRID = "anime_hybrid"
    ANIME_GUOFENG = "anime_guofeng"
    ANIME_ORANGE = "anime_orange"
    DREAMLIKE = "dreamlike_v1"


def get_sd_tag(version: SDVersions) -> str:
    if version == SDVersions.v1_5_BC:
        return "v1.5"
    if version == SDVersions.ANIME:
        return "anime_nai"
    if version == SDVersions.ANIME_ANYTHING:
        return "anime_anything_v3"
    if version == SDVersions.ANIME_HYBRID:
        return "anime_hybrid_v1"
    if version == SDVersions.ANIME_GUOFENG:
        return "anime_guofeng3"
    if version == SDVersions.ANIME_ORANGE:
        return "anime_orange2"
    return version


def get_txt_cond(t: Union[str, List[str]], n: Optional[int]) -> Tuple[List[str], int]:
    if n is None:
        n = 1 if isinstance(t, str) else len(t)
    if isinstance(t, str):
        t = [t] * n
    if len(t) != n:
        raise ValueError(
            f"`num_samples` ({n}) should be identical with "
            f"the number of `txt` ({len(t)})"
        )
    return t, n


def get_size(
    size: Optional[Tuple[int, int]],
    anchor: int,
    max_wh: int,
) -> Optional[Tuple[int, int]]:
    if size is None:
        return None
    new_size = restrict_wh(*size, max_wh)
    return tuple(map(get_suitable_size, new_size, (anchor, anchor)))  # type: ignore


class DiffusionAPI(APIMixin):
    m: DDPM
    sampler: ISampler
    cond_model: Optional[nn.Module]
    first_stage: Optional[IAutoEncoder]
    latest_seed: int
    latest_variation_seed: Optional[int]
    sd_weights: WeightsPool
    current_sd_version: Optional[SDVersions]

    def __init__(
        self,
        m: DDPM,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
        clip_skip: int = 0,
    ):
        super().__init__(m, device, use_amp=use_amp, use_half=use_half)
        self.sampler = m.sampler
        self.cond_type = m.condition_type
        self.clip_skip = clip_skip
        self.sd_weights = WeightsPool()
        self.current_sd_version = None
        # extracted the condition model so we can pre-calculate the conditions
        self.cond_model = m.condition_model
        if self.cond_model is not None:
            self.cond_model.eval()
        m.condition_model = nn.Identity()
        # pre-calculate unconditional_cond if needed
        self._original_raw_uncond = getattr(m.sampler, "unconditional_cond", None)
        self._uncond_cache: tensor_dict_type = {}
        self._update_sampler_uncond(clip_skip)
        # extract first stage
        if not isinstance(m, LDM):
            self.first_stage = None
        else:
            self.first_stage = m.first_stage

    # api

    @property
    def size_info(self) -> SizeInfo:
        opt_size = self.m.img_size
        if self.first_stage is None:
            factor = 1
        else:
            factor = self.first_stage.img_size // opt_size
        return SizeInfo(factor, opt_size)

    def to(
        self,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> None:
        super().to(device, use_amp=use_amp, use_half=use_half)
        unconditional_cond = getattr(self.sampler, "unconditional_cond", None)
        if use_half:
            if self.cond_model is not None:
                self.cond_model.half()
            if self.first_stage is not None:
                self.first_stage.half()
            if unconditional_cond is not None:
                unconditional_cond = unconditional_cond.half()
            for k, v in self._uncond_cache.items():
                self._uncond_cache[k] = v.half()
        else:
            if self.cond_model is not None:
                self.cond_model.float()
            if self.first_stage is not None:
                self.first_stage.float()
            if unconditional_cond is not None:
                unconditional_cond = unconditional_cond.float()
            for k, v in self._uncond_cache.items():
                self._uncond_cache[k] = v.float()
        if self.cond_model is not None:
            self.cond_model.to(device)
        if self.first_stage is not None:
            self.first_stage.to(device)
        if unconditional_cond is not None:
            self.sampler.unconditional_cond = unconditional_cond.to(device)
        for k, v in self._uncond_cache.items():
            self._uncond_cache[k] = v.to(device)

    def prepare_sd(self, versions: List[SDVersions]) -> None:
        root = os.path.join(OPT.cache_dir, DLZoo.model_dir)
        for tag in map(get_sd_tag, versions):
            if tag not in self.sd_weights:
                model_path = download_model(f"ldm_sd_{tag}", root=root)
                self.sd_weights.register(tag, model_path)

    def switch_sd(self, version: SDVersions) -> None:
        tag = get_sd_tag(version)
        if self.current_sd_version is not None:
            if tag == get_sd_tag(self.current_sd_version):
                return
        d = self.sd_weights.get(tag)
        with self.load_context() as m:
            m.load_state_dict(d)
        self.current_sd_version = version

    def get_cond(self, cond: Any) -> Tensor:
        if self.cond_model is None:
            msg = "should not call `get_cond` when `cond_model` is not available"
            raise ValueError(msg)
        with torch.no_grad():
            with self.amp_context:
                return self.cond_model(cond)

    def switch_sampler(
        self,
        sampler: str,
        sampler_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        sampler_ins = self.m.make_sampler(sampler, sampler_config)
        current_unconditional_cond = getattr(self.m.sampler, "unconditional_cond", None)
        if current_unconditional_cond is not None:
            if hasattr(sampler_ins, "unconditional_cond"):
                sampler_ins.unconditional_cond = current_unconditional_cond
        current_guidance = getattr(self.m.sampler, "unconditional_guidance_scale", None)
        if current_guidance is not None:
            if hasattr(sampler_ins, "unconditional_guidance_scale"):
                sampler_ins.unconditional_guidance_scale = current_guidance
        self.sampler = self.m.sampler = sampler_ins

    def switch_circular(self, enable: bool) -> None:
        def _inject(m: nn.Module) -> None:
            for child in m.children():
                _inject(child)
            modules.append(m)

        padding_mode = "circular" if enable else "zeros"
        modules: List[nn.Module] = []
        _inject(self.m)
        for module in modules:
            if isinstance(module, nn.Conv2d):
                module.padding_mode = padding_mode
            elif isinstance(module, Conv2d):
                module.padding = padding_mode

    def sample(
        self,
        num_samples: int,
        export_path: Optional[str] = None,
        *,
        seed: Optional[int] = None,
        use_seed_resize: bool = False,
        # each variation contains (seed, weight)
        variations: Optional[List[Tuple[int, float]]] = None,
        variation_seed: Optional[int] = None,
        variation_strength: Optional[float] = None,
        z: Optional[Tensor] = None,
        z_ref: Optional[Tensor] = None,
        z_ref_mask: Optional[Tensor] = None,
        z_ref_noise: Optional[Tensor] = None,
        size: Optional[Tuple[int, int]] = None,
        original_size: Optional[Tuple[int, int]] = None,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        cond_concat: Optional[Tensor] = None,
        unconditional_cond: Optional[Any] = None,
        hint: Optional[Union[Tensor, tensor_dict_type]] = None,
        hint_start: Optional[Union[float, Dict[str, float]]] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        batch_size: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if batch_size is None:
            batch_size = num_samples
        registered_custom = False
        if self.cond_model is not None:
            clip_skip = kwargs.get(
                "clip_skip",
                0 if self.clip_skip is None else self.clip_skip,
            )
            self._update_sampler_uncond(clip_skip)
            if isinstance(self.cond_model, CLIPTextConditionModel):
                custom_embeddings = kwargs.get("custom_embeddings")
                if custom_embeddings is not None:
                    registered_custom = True
                    self.cond_model.register_custom(custom_embeddings)
        if cond is not None:
            if self.cond_type != CONCAT_TYPE and self.cond_model is not None:
                cond = predict_array_data(
                    self.cond_model,
                    ArrayData.init().fit(np.array(cond)),
                    batch_size=batch_size,
                )[PREDICTIONS_KEY]
        if cond is not None and num_samples != len(cond):
            raise ValueError(
                f"`num_samples` ({num_samples}) should be identical with "
                f"the number of `cond` ({len(cond)})"
            )
        if alpha is not None and original_size is not None:
            alpha_h, alpha_w = alpha.shape[-2:]
            if alpha_w != original_size[0] or alpha_h != original_size[1]:
                raise ValueError(
                    f"shape of the provided `alpha` ({alpha_w}, {alpha_h}) should be "
                    f"identical with the provided `original_size` {original_size}"
                )
        unconditional = cond is None
        if unconditional:
            cond = [0] * num_samples
        cond_data: ArrayData = ArrayData.init(DataConfig(batch_size=batch_size))
        cond_data.fit(cond)
        iterator = TensorBatcher(cond_data.get_loaders()[0], self.device)
        num_iter = len(iterator)
        if verbose and num_iter > 1:
            iterator = tqdm(iterator, desc="iter", total=num_iter)
        sampled = []
        kw = dict(num_steps=num_steps, verbose=verbose)
        kw.update(shallow_copy_dict(kwargs))
        factor, opt_size = self.size_info
        if size is None:
            size = opt_size, opt_size
        else:
            size = tuple(map(lambda n: round(n / factor), size))  # type: ignore
        uncond_backup = None
        unconditional_cond_backup = None
        if self.cond_model is not None and unconditional_cond is not None:
            uncond_backup = getattr(self.sampler, "uncond", None)
            unconditional_cond_backup = getattr(
                self.sampler,
                "unconditional_cond",
                None,
            )
            uncond = self.get_cond(unconditional_cond).to(self.device)
            self.sampler.uncond = uncond.clone()
            self.sampler.unconditional_cond = uncond.clone()
        with eval_context(self.m):
            with self.amp_context:
                for i, batch in enumerate(iterator):
                    # from the 2nd batch forward, we need to re-generate new seeds
                    if i >= 1:
                        seed = new_seed()
                    i_kw = shallow_copy_dict(kw)
                    i_cond = batch[INPUT_KEY].to(self.device)
                    i_n = len(i_cond)
                    repeat = (
                        lambda t: t
                        if t.shape[0] == i_n
                        else t.repeat_interleave(i_n, dim=0)
                    )
                    if z is not None:
                        i_z = repeat(z)
                    else:
                        in_channels = self.m.in_channels
                        if self.cond_type == CONCAT_TYPE:
                            in_channels -= cond.shape[1]
                        elif cond_concat is not None:
                            in_channels -= cond_concat.shape[1]
                        i_z_shape = i_n, in_channels, *size[::-1]
                        i_z, _ = self._set_seed_and_variations(
                            seed,
                            lambda: torch.randn(i_z_shape, device=self.device),
                            lambda noise: noise,
                            variations,
                            variation_seed,
                            variation_strength,
                        )
                    if use_seed_resize:
                        z_original_shape = list(i_z.shape[-2:])
                        z_opt_shape = list(
                            map(lambda n: round(n / factor), [opt_size, opt_size])
                        )
                        if z_original_shape != z_opt_shape:
                            dx = (z_original_shape[0] - z_opt_shape[0]) // 2
                            dy = (z_original_shape[1] - z_opt_shape[1]) // 2
                            x = z_opt_shape[0] if dx >= 0 else z_opt_shape[0] + 2 * dx
                            y = z_opt_shape[1] if dy >= 0 else z_opt_shape[1] + 2 * dy
                            dx = max(-dx, 0)
                            dy = max(-dy, 0)
                            i_opt_z_shape = (
                                i_n,
                                self.m.in_channels,
                                *z_opt_shape,
                            )
                            i_opt_z, _ = self._set_seed_and_variations(
                                seed,
                                lambda: torch.randn(i_opt_z_shape, device=self.device),
                                lambda noise: noise,
                                variations,
                                variation_seed,
                                variation_strength,
                            )
                            i_z[..., dx : dx + x, dy : dy + y] = i_opt_z[
                                ..., dx : dx + x, dy : dy + y
                            ]
                    if z_ref is not None and z_ref_mask is not None:
                        if z_ref_noise is not None:
                            i_kw["ref"] = repeat(z_ref)
                            i_kw["ref_mask"] = repeat(z_ref_mask)
                            i_kw["ref_noise"] = repeat(z_ref_noise)
                    if unconditional:
                        i_cond = None
                    if self.use_half:
                        i_z = i_z.half()
                        if i_cond is not None:
                            i_cond = i_cond.half()
                        for k, v in i_kw.items():
                            if isinstance(v, torch.Tensor) and v.is_floating_point():
                                i_kw[k] = v.half()
                    if cond_concat is not None:
                        if self.cond_type != HYBRID_TYPE:
                            raise ValueError(
                                f"condition type should be `{HYBRID_TYPE}` when "
                                f"`cond_concat` is provided"
                            )
                        i_cond = {
                            CROSS_ATTN_KEY: i_cond,
                            CONCAT_KEY: cond_concat,
                        }
                    if hint is not None:
                        if isinstance(i_cond, dict):
                            i_cond[CONTROL_HINT_KEY] = hint
                            i_cond[CONTROL_HINT_START_KEY] = hint_start
                        else:
                            i_cond = {
                                CROSS_ATTN_KEY: i_cond,
                                CONTROL_HINT_KEY: hint,
                                CONTROL_HINT_START_KEY: hint_start,
                            }
                    with switch_sampler_context(self, i_kw.get("sampler")):
                        i_sampled = self.m.decode(i_z, cond=i_cond, **i_kw)
                    sampled.append(i_sampled.cpu().float())
        if uncond_backup is not None:
            self.sampler.uncond = uncond_backup
        if unconditional_cond_backup is not None:
            self.sampler.unconditional_cond = unconditional_cond_backup
        concat = torch.cat(sampled, dim=0)
        if clip_output:
            concat = torch.clip(concat, -1.0, 1.0)
        if callback is not None:
            concat = callback(concat)
        if original_size is not None:
            original_size = (max(original_size[0], 1), max(original_size[1], 1))
            with torch.no_grad():
                concat = F.interpolate(
                    concat,
                    original_size[::-1],
                    mode="bicubic",
                )
        if alpha is not None:
            alpha = torch.from_numpy(2.0 * alpha - 1.0)
            if original_size is None:
                with torch.no_grad():
                    alpha = F.interpolate(
                        alpha,
                        concat.shape[-2:],
                        mode="nearest",
                    )
            concat = torch.cat([concat, alpha], dim=1)
        if export_path is not None:
            save_images(concat, export_path)
        self.empty_cuda_cache()
        if registered_custom:
            self.cond_model.clear_custom()
        return concat

    def txt2img(
        self,
        txt: Union[str, List[str]],
        export_path: Optional[str] = None,
        *,
        anchor: int = 64,
        max_wh: int = 512,
        num_samples: Optional[int] = None,
        size: Optional[Tuple[int, int]] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        batch_size: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        txt, num_samples = get_txt_cond(txt, num_samples)
        new_size = get_size(size, anchor, max_wh)
        return self.sample(
            num_samples,
            export_path,
            size=new_size,
            original_size=size,
            cond=txt,
            num_steps=num_steps,
            clip_output=clip_output,
            callback=callback,
            batch_size=batch_size,
            verbose=verbose,
            **kwargs,
        )

    def txt2img_inpainting(
        self,
        txt: Union[str, List[str]],
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        seed: Optional[int] = None,
        reference: Optional[Union[str, Image.Image]] = None,
        reference_fidelity: float = 0.2,
        anchor: int = 64,
        max_wh: int = 512,
        num_samples: Optional[int] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        keep_original: bool = False,
        use_raw_inpainting: bool = False,
        raw_inpainting_fidelity: float = 0.2,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        use_latent_guidance: bool = False,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if use_raw_inpainting:
            image_res = read_image(image, max_wh, anchor=anchor)
            mask = read_image(mask, max_wh, anchor=anchor, to_mask=True).image
            z = z_ref = self._get_z(image_res.image)
            z_ref_mask = F.interpolate(
                torch.from_numpy(mask).to(z_ref),
                z_ref.shape[-2:],
                mode="bicubic",
            )
            return self._img2img(
                z,
                export_path,
                z_ref=z_ref,
                z_ref_mask=z_ref_mask,
                fidelity=raw_inpainting_fidelity,
                original_size=image_res.original_size,
                alpha=None,
                cond=[txt],
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )
        txt_list, num_samples = get_txt_cond(txt, num_samples)
        res = self._get_masked_cond(
            image,
            mask,
            max_wh,
            anchor,
            lambda remained_mask, img: np.where(remained_mask, img, 0.5),
            lambda bool_mask: torch.from_numpy(bool_mask),
        )
        # sampling
        with switch_sampler_context(self, kwargs.get("sampler")):
            # calculate `z_ref` stuffs based on `use_image_guidance`
            if not use_latent_guidance:
                z_ref = z_ref_mask = z_ref_noise = None
            else:
                z_ref = self._get_z(res.image_res.image)
                z_ref_mask = 1.0 - F.interpolate(
                    torch.from_numpy(res.mask).to(z_ref),
                    z_ref.shape[-2:],
                    mode="bicubic",
                )
                if seed is not None:
                    seed_everything(seed)
                z_ref_noise = torch.randn_like(z_ref)
                reference = res.image_res.original
            # calculate `z` based on `reference`
            if reference is None:
                z = None
                size = tuple(
                    map(
                        lambda n: n * self.size_info.factor,
                        res.remained_image_cond.shape[-2:][::-1],
                    )
                )
            else:
                size = None
                z = self._get_z(read_image(reference, max_wh, anchor=anchor).image)
                if z_ref_mask is not None and z_ref_noise is not None:
                    z = z * z_ref_mask + z_ref_noise * (1.0 - z_ref_mask)
                z, _, kwargs = self._q_sample(
                    z,
                    num_steps,
                    reference_fidelity,
                    seed,
                    **kwargs,
                )
            # core
            sampled = self.sample(
                num_samples,
                export_path,
                seed=seed,
                z=z,
                z_ref=z_ref,
                z_ref_mask=z_ref_mask,
                z_ref_noise=z_ref_noise,
                size=size,  # type: ignore
                original_size=res.image_res.original_size,
                alpha=res.image_res.alpha,
                cond=txt_list,
                cond_concat=torch.cat([res.mask_cond, res.remained_image_cond], dim=1),
                num_steps=num_steps,
                clip_output=clip_output,
                callback=callback,
                verbose=verbose,
                **kwargs,
            )
        if keep_original:
            original = np.array(res.image_res.original).astype(np.float32) / 127.5 - 1.0
            original = original.transpose([2, 0, 1])[None]
            mask_ = read_image(res.mask_res.original, None, anchor=None, to_mask=True)
            remained_mask = ~(mask_.image >= 0.5)
            sampled = np.where(remained_mask, original, sampled.numpy())
            sampled = torch.from_numpy(sampled)
        return sampled

    def outpainting(
        self,
        txt: str,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 64,
        max_wh: int = 512,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        keep_original: bool = False,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if isinstance(image, str):
            image = Image.open(image)
        if image.mode != "RGBA":
            raise ValueError("`image` should be `RGBA` in outpainting")
        *rgb, alpha = image.split()
        mask = Image.fromarray(255 - np.array(alpha))
        image = Image.merge("RGB", rgb)
        return self.txt2img_inpainting(
            txt,
            image,
            mask,
            export_path,
            anchor=anchor,
            max_wh=max_wh,
            num_steps=num_steps,
            clip_output=clip_output,
            keep_original=keep_original,
            callback=callback,
            verbose=verbose,
            **kwargs,
        )

    def img2img(
        self,
        image: Union[str, Image.Image, Tensor],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        fidelity: float = 0.2,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if isinstance(image, Tensor):
            original_size = tuple(image.shape[-2:][::-1])
        else:
            res = read_image(image, max_wh, anchor=anchor)
            image = res.image
            original_size = res.original_size
            if alpha is None:
                alpha = res.alpha
        z = self._get_z(image)
        return self._img2img(
            z,
            export_path,
            fidelity=fidelity,
            original_size=original_size,  # type: ignore
            alpha=alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def inpainting(
        self,
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        refine_fidelity: Optional[float] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        # inpainting callback, will not trigger in refine stage
        def callback(out: Tensor) -> Tensor:
            final = torch.from_numpy(res.remained_image.copy())
            final += 0.5 * (1.0 + out) * (1.0 - res.remained_mask)
            return 2.0 * final - 1.0

        res = self._get_masked_cond(
            image,
            mask,
            max_wh,
            anchor,
            lambda remained_mask, img: np.where(remained_mask, img, 0.0),
            lambda bool_mask: torch.where(torch.from_numpy(bool_mask), 1.0, -1.0),
        )
        cond = torch.cat([res.remained_image_cond, res.mask_cond], dim=1)
        size = self._get_identical_size_with(res.remained_image_cond)
        # refine with img2img
        if refine_fidelity is not None:
            z = self._get_z(res.image_res.image)
            return self._img2img(
                z,
                export_path,
                fidelity=refine_fidelity,
                original_size=res.image_res.original_size,
                alpha=res.image_res.alpha if alpha is None else alpha,
                cond=cond,
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )
        # sampling
        return self.sample(
            1,
            export_path,
            size=size,  # type: ignore
            original_size=res.image_res.original_size,
            alpha=res.image_res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            callback=callback,
            verbose=verbose,
            **kwargs,
        )

    def sr(
        self,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 8,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if not isinstance(self.m, LDM):
            raise ValueError("`sr` is now only available for `LDM` models")
        factor = 2 ** (len(self.m.first_stage.core.channel_multipliers) - 1)
        res = read_image(image, round(max_wh / factor), anchor=anchor)
        wh_ratio = res.original_size[0] / res.original_size[1]
        zh, zw = res.image.shape[-2:]
        sr_size = (zw, zw / wh_ratio) if zw > zh else (zh * wh_ratio, zh)
        sr_size = tuple(map(lambda n: round(factor * n), sr_size))  # type: ignore
        cond = torch.from_numpy(2.0 * res.image - 1.0).to(self.device)
        size = self._get_identical_size_with(cond)
        return self.sample(
            1,
            export_path,
            size=size,
            original_size=sr_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def semantic2img(
        self,
        semantic: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 16,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        err_fmt = "`{}` is needed for `semantic2img`"
        if self.cond_model is None:
            raise ValueError(err_fmt.format("cond_model"))
        in_channels = getattr(self.cond_model, "in_channels", None)
        if in_channels is None:
            raise ValueError(err_fmt.format("cond_model.in_channels"))
        factor = getattr(self.cond_model, "factor", None)
        if factor is None:
            raise ValueError(err_fmt.format("cond_model.factor"))
        res = read_image(
            semantic,
            max_wh,
            anchor=anchor,
            to_gray=True,
            resample=Image.NEAREST,
            normalize=False,
        )
        cond = torch.from_numpy(res.image).to(torch.long).to(self.device)
        cond = F.one_hot(cond, num_classes=in_channels)[0]
        cond = cond.half() if self.use_half else cond.float()
        cond = cond.permute(0, 3, 1, 2).contiguous()
        cond = self.get_cond(cond)
        size = self._get_identical_size_with(cond)
        return self.sample(
            1,
            export_path,
            size=size,
            original_size=res.original_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def load_context(self, *, ignore_lora: bool = True) -> ContextManager:
        class _:
            def __init__(self, api: DiffusionAPI):
                self.api = api
                self.m_ctrl = api.m.control_model
                self.m_cond = api.m.condition_model
                api.m.control_model = None
                api.m.condition_model = api.cond_model
                if not ignore_lora:
                    self.lora_checkpoints = None
                else:
                    if not isinstance(api.m, StableDiffusion):
                        msg = "currently only `StableDiffusion` supports `ignore_lora`"
                        raise ValueError(msg)
                    if not api.m.has_lora:
                        self.lora_checkpoints = None
                    else:
                        self.lora_checkpoints = api.m.get_lora_checkpoints()
                        api.m.cleanup_lora()

            def __enter__(self) -> DDPM:
                return self.api.m

            def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                if self.lora_checkpoints is not None:
                    assert isinstance(self.api.m, StableDiffusion)
                    self.api.m.restore_lora_from(self.lora_checkpoints)
                self.api.m.control_model = self.m_ctrl
                self.api.m.condition_model = self.m_cond

        return _(self)

    # lora

    def load_sd_lora(self, key: str, *, path: str) -> None:
        if not isinstance(self.m, StableDiffusion):
            raise ValueError("only `StableDiffusion` can use `load_sd_lora`")
        with self.load_context(ignore_lora=False):
            self.m.load_lora(key, path=path)

    def inject_sd_lora(self, *keys: str) -> None:
        if not isinstance(self.m, StableDiffusion):
            raise ValueError("only `StableDiffusion` can use `inject_sd_lora`")
        with self.load_context(ignore_lora=False):
            self.m.inject_lora(*keys)

    def cleanup_sd_lora(self) -> None:
        if not isinstance(self.m, StableDiffusion):
            raise ValueError("only `StableDiffusion` can use `cleanup_sd_lora`")
        with self.load_context(ignore_lora=False):
            self.m.cleanup_lora()

    def set_sd_lora_scales(self, scales: Dict[str, float]) -> None:
        if not isinstance(self.m, StableDiffusion):
            raise ValueError("only `StableDiffusion` can use `set_sd_lora_scales`")
        with self.load_context(ignore_lora=False):
            self.m.set_lora_scales(scales)

    # tomesd

    def set_tome_info(self, tome_info: Optional[Dict[str, Any]]) -> None:
        self.m.unet.set_tome_info(tome_info)

    # constructors

    @classmethod
    def from_sd(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        return cls.from_pipeline(ldm_sd(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_version(
        cls: Type[T],
        version: SDVersions,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        m = ldm_sd_tag(get_sd_tag(version))
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_anime(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        v = SDVersions.ANIME
        return cls.from_sd_version(v, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_inpainting(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        m = ldm_sd_inpainting()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_v2(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        return cls.from_pipeline(
            ldm_sd_v2(),
            device,
            use_amp=use_amp,
            use_half=use_half,
            clip_skip=1,
        )

    @classmethod
    def from_sd_v2_base(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        return cls.from_pipeline(
            ldm_sd_v2_base(),
            device,
            use_amp=use_amp,
            use_half=use_half,
            clip_skip=1,
        )

    @classmethod
    def from_celeba_hq(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        m = ldm_celeba_hq()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_inpainting(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        m = ldm_inpainting()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sr(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        return cls.from_pipeline(ldm_sr(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_semantic(
        cls: Type[T],
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        m = ldm_semantic()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    # internal

    def _get_z(self, img: arr_type) -> Tensor:
        img = 2.0 * img - 1.0
        z = img if isinstance(img, Tensor) else torch.from_numpy(img)
        if self.use_half:
            z = z.half()
        z = z.to(self.device)
        z = self.m._preprocess(z, deterministic=True)
        return z

    def _get_identical_size_with(self, pivot: Tensor) -> Tuple[int, int]:
        return tuple(  # type: ignore
            map(
                lambda n: n * self.size_info.factor,
                pivot.shape[-2:][::-1],
            )
        )

    def _set_seed_and_variations(
        self,
        seed: Optional[int],
        get_noise: Callable[[], Tensor],
        get_new_z: Callable[[Tensor], Tensor],
        variations: Optional[List[Tuple[int, float]]],
        variation_seed: Optional[int],
        variation_strength: Optional[float],
    ) -> Tuple[Tensor, Tensor]:
        if seed is None:
            seed = new_seed()
        seed = seed_everything(seed)
        self.latest_seed = seed
        z_noise = get_noise()
        self.latest_variation_seed = None
        if variations is not None:
            for v_seed, v_weight in variations:
                seed_everything(v_seed)
                v_noise = get_noise()
                z_noise = slerp(v_noise, z_noise, v_weight)
        if variation_strength is not None:
            random.seed()
            if variation_seed is None:
                variation_seed = new_seed()
            variation_seed = seed_everything(variation_seed)
            self.latest_variation_seed = variation_seed
            variation_noise = get_noise()
            z_noise = slerp(variation_noise, z_noise, variation_strength)
        z = get_new_z(z_noise)
        return z, z_noise

    def _update_clip_skip(self, clip_skip: int) -> None:
        if isinstance(self.cond_model, CLIPTextConditionModel):
            self.cond_model.clip_skip = clip_skip

    def _update_sampler_uncond(self, clip_skip: int) -> None:
        self._update_clip_skip(clip_skip)
        if self.cond_model is not None and self._original_raw_uncond is not None:
            cache = self._uncond_cache.get(clip_skip)
            if cache is not None:
                uncond = cache
            else:
                uncond = self.get_cond(self._original_raw_uncond)
                self._uncond_cache[clip_skip] = uncond
            self.m.sampler.unconditional_cond = uncond.to(self.device)

    def _get_masked_cond(
        self,
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        max_wh: int,
        anchor: int,
        mask_image_fn: Callable[[np.ndarray, np.ndarray], np.ndarray],
        mask_cond_fn: Callable[[np.ndarray], Tensor],
    ) -> MaskedCond:
        # handle mask stuffs
        image_res = read_image(image, max_wh, anchor=anchor)
        mask_res = read_image(mask, max_wh, anchor=anchor, to_mask=True)
        mask = mask_res.image
        bool_mask = np.round(mask) >= 0.5
        remained_mask = (~bool_mask).astype(np.float16 if self.use_half else np.float32)
        remained_image = mask_image_fn(remained_mask, image_res.image)
        # construct condition tensor
        remained_cond = self._get_z(remained_image)
        latent_shape = remained_cond.shape[-2:]
        mask_cond = mask_cond_fn(bool_mask).to(torch.float32)
        mask_cond = F.interpolate(mask_cond, size=latent_shape)
        if self.use_half:
            mask_cond = mask_cond.half()
        mask_cond = mask_cond.to(self.device)
        return MaskedCond(
            image_res,
            mask_res,
            mask,
            remained_image,
            remained_mask,
            mask_cond,
            remained_cond,
        )

    def _q_sample(
        self,
        z: Tensor,
        num_steps: Optional[int],
        fidelity: float,
        seed: Optional[int],
        variations: Optional[List[Tuple[int, float]]] = None,
        variation_seed: Optional[int] = None,
        variation_strength: Optional[float] = None,
        **kwargs: Any,
    ) -> Tuple[Tensor, Tensor, Dict[str, Any]]:
        if num_steps is None:
            num_steps = self.sampler.default_steps
        t = min(num_steps, round((1.0 - fidelity) * (num_steps + 1)))
        ts = get_timesteps(t, 1, z.device)
        if isinstance(self.sampler, (DDIMMixin, KSamplerMixin, DPMSolver)):
            kw = shallow_copy_dict(self.sampler.sample_kwargs)
            kw["total_step"] = num_steps
            safe_execute(self.sampler._reset_buffers, kw)
        z, noise = self._set_seed_and_variations(
            seed,
            lambda: torch.randn_like(z),
            lambda noise_: self.sampler.q_sample(z, ts, noise_),
            variations,
            variation_seed,
            variation_strength,
        )
        kwargs["start_step"] = num_steps - t
        return z, noise, kwargs

    def _img2img(
        self,
        z: Tensor,
        export_path: Optional[str] = None,
        *,
        z_ref: Optional[Tensor] = None,
        z_ref_mask: Optional[Tensor] = None,
        original_size: Optional[Tuple[int, int]] = None,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        with switch_sampler_context(self, kwargs.get("sampler")):
            z, noise, kwargs = self._q_sample(z, num_steps, **kwargs)
            return self.sample(
                z.shape[0],
                export_path,
                z=z,
                z_ref=z_ref,
                z_ref_mask=z_ref_mask,
                z_ref_noise=None if z_ref is None else noise,
                original_size=original_size,
                alpha=alpha,
                cond=cond,
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )


def offset_cnet_weights(
    d: tensor_dict_type,
    *,
    api: Optional[DiffusionAPI] = None,
    base_md: Optional[tensor_dict_type] = None,
) -> tensor_dict_type:
    with open(download_static("sd_mapping", extension="json"), "r") as f:
        mapping = json.load(f)
    with open(download_static("sd_controlnet_mapping", extension="json"), "r") as f:
        c_mapping = json.load(f)
    rev_c_mapping = {v: k for k, v in c_mapping.items()}
    nd = shallow_copy_dict(d)
    if base_md is None:
        if api is None:
            raise ValueError("Either `api` or `md` must be provided.")
        with api.load_context() as m:
            base_md = m.state_dict()
    device = list(base_md.values())[0].device
    nd = {k: v.to(device) for k, v in nd.items()}
    for k, v in nd.items():
        rev_k = rev_c_mapping[k]
        original_k = f"model.diffusion_model.{rev_k.split('.', 1)[1]}"
        mk = mapping.get(original_k)
        if mk is None:
            continue
        mv = base_md[mk].to(v)
        # inpainting workaround
        if k == "input_blocks.0.0.weight" and mv.shape[1] == 9:
            mv = mv[:, :4]
        nd[k] = v + mv
    return nd


class ControlNetHints(str, Enum):
    DEPTH = "depth"
    CANNY = "canny"
    POSE = "pose"
    MLSD = "mlsd"


class Annotator(ABC):
    @abstractmethod
    def __init__(self, device: torch.device) -> None:
        pass

    @abstractmethod
    def to(self: TAnnotator, device: torch.device, *, use_half: bool) -> TAnnotator:
        pass

    @abstractmethod
    def annotate(self, uint8_rgb: np.ndarray, **kwargs: Any) -> np.ndarray:
        pass


class DepthAnnotator(Annotator):
    def __init__(self, device: torch.device) -> None:
        self.m = MiDaSAPI(device)

    def to(self, device: torch.device, *, use_half: bool) -> "DepthAnnotator":
        self.m.to(device, use_half=use_half)
        return self

    def annotate(self, uint8_rgb: np.ndarray) -> np.ndarray:  # type: ignore
        return self.m.detect_depth(uint8_rgb)


class CannyAnnotator(Annotator):
    def __init__(self, device: torch.device) -> None:
        if cv2 is None:
            raise ValueError("`cv2` is needed for `CannyAnnotator`")

    def to(self, device: torch.device, *, use_half: bool) -> "CannyAnnotator":
        return self

    def annotate(  # type: ignore
        self,
        uint8_rgb: np.ndarray,
        *,
        low_threshold: int,
        high_threshold: int,
    ) -> np.ndarray:
        return cv2.Canny(uint8_rgb, low_threshold, high_threshold)


class PoseAnnotator(Annotator):
    def __init__(self, device: torch.device) -> None:
        self.m = OpenposeDetector(device)

    def to(self, device: torch.device, *, use_half: bool) -> "PoseAnnotator":
        self.m.to(device, use_half=use_half)
        return self

    def annotate(self, uint8_rgb: np.ndarray) -> np.ndarray:  # type: ignore
        return self.m(uint8_rgb)[0]


class MLSDAnnotator(Annotator):
    def __init__(self, device: torch.device) -> None:
        self.m = MLSDDetector(device)

    def to(self, device: torch.device, *, use_half: bool) -> "MLSDAnnotator":
        self.m.to(device, use_half=use_half)
        return self

    def annotate(  # type: ignore
        self,
        uint8_rgb: np.ndarray,
        *,
        value_threshold: float,
        distance_threshold: float,
    ) -> np.ndarray:
        return self.m(uint8_rgb, value_threshold, distance_threshold)


class ControlledDiffusionAPI(DiffusionAPI):
    loaded: Dict[ControlNetHints, bool]
    annotators: Dict[ControlNetHints, Annotator]
    base_sd_versions: Dict[ControlNetHints, SDVersions]
    controlnet_weights: Dict[ControlNetHints, tensor_dict_type]

    control_defaults = {
        ControlNetHints.DEPTH: "ldm.sd_v1.5.control.diff.depth",
        ControlNetHints.CANNY: "ldm.sd_v1.5.control.diff.canny",
        ControlNetHints.POSE: "ldm.sd_v1.5.control.diff.pose",
        ControlNetHints.MLSD: "ldm.sd_v1.5.control.diff.mlsd",
    }
    annotator_classes: Dict[ControlNetHints, Type[Annotator]] = {
        ControlNetHints.DEPTH: DepthAnnotator,
        ControlNetHints.CANNY: CannyAnnotator,
        ControlNetHints.POSE: PoseAnnotator,
        ControlNetHints.MLSD: MLSDAnnotator,
    }

    def __init__(
        self,
        m: DDPM,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
        clip_skip: int = 0,
        hint_channels: int = 3,
        num_pool: int = 4,
    ):
        super().__init__(
            m,
            device,
            use_amp=use_amp,
            use_half=use_half,
            clip_skip=clip_skip,
        )
        pool = sorted(self.control_defaults)
        selected_pool = pool[: min(num_pool, len(pool))]
        self.m.make_control_net({k: hint_channels for k in selected_pool})
        self.loaded = {k: False for k in selected_pool}
        self.annotators = {}
        self.num_pool = num_pool
        self.control_model = self.m.control_model
        self.base_sd_versions = {}
        self.controlnet_weights = {}

    def to(
        self,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
        no_annotator: bool = False,
    ) -> None:
        super().to(device, use_amp=use_amp, use_half=use_half)
        if not no_annotator:
            for annotator in self.annotators.values():
                annotator = annotator.to(device, use_half=use_half)

    @property
    def available_control_hints(self) -> List[ControlNetHints]:
        return list(self.controlnet_weights)

    def set_tome_info(self, tome_info: Optional[Dict[str, Any]]) -> None:
        super().set_tome_info(tome_info)
        if self.control_model is not None:
            if isinstance(self.control_model, ControlNet):
                self.control_model.set_tome_info(tome_info)
            else:
                for m in self.control_model.values():
                    m.set_tome_info(tome_info)

    def prepare_control(self, hints2tags: Dict[ControlNetHints, str]) -> None:
        root = os.path.join(OPT.cache_dir, DLZoo.model_dir)
        for hint, tag in hints2tags.items():
            self.controlnet_weights[hint] = torch.load(download_model(tag, root=root))

    def prepare_control_defaults(self) -> None:
        self.prepare_control(self.control_defaults)

    def switch_control(
        self,
        *hints: ControlNetHints,
        base_md: Optional[tensor_dict_type] = None,
    ) -> None:
        if self.m.control_model is None:
            raise ValueError("`control_model` is not built yet")

        hints_list = list(hints)
        if len(hints_list) > self.num_pool:
            print_warning(
                f"number of target hints ({len(hints_list)}) exceeds "
                f"number of pool ({self.num_pool}), "
                f"so only {self.num_pool} hints will be activated"
            )
            random.shuffle(hints_list)
            hints_list = hints_list[: self.num_pool]

        target = set(hints_list)
        current = set(self.loaded)
        not_current = sorted(target - current)
        if not_current:
            not_target = sorted(current - target)
            for i, i_not_current in enumerate(not_current):
                pop_key = not_target[i]
                self.m.rename_control_net(pop_key, i_not_current)
                self.loaded[i_not_current] = False
                self.loaded.pop(pop_key)

        sorted_target = sorted(target)
        loaded_list = [self.loaded[hint] for hint in sorted_target]
        base_list = [self.base_sd_versions.get(hint) for hint in sorted_target]
        need_offset_list = [
            base is None
            or self.current_sd_version is None
            or base_md is not None
            or get_sd_tag(base) != get_sd_tag(self.current_sd_version)
            for base in base_list
        ]
        if all(loaded_list) and not any(need_offset_list):
            return
        iterator = zip(loaded_list, sorted_target, need_offset_list)
        for loaded, hint, need_offset in iterator:
            if loaded and not need_offset:
                continue
            d = self.controlnet_weights.get(hint)
            if d is None:
                raise ValueError(
                    f"cannot find ControlNet weights called '{hint}', "
                    f"available weights are: {', '.join(self.available_control_hints)}"
                )
            if need_offset:
                d = offset_cnet_weights(d, api=self, base_md=base_md)
            self.m.load_control_net_with(hint, d)
            self.loaded[hint] = True
            # if `base_md` is provided, should reset cache settings
            if base_md is not None:
                self.base_sd_versions.pop(hint, None)
            else:
                if self.current_sd_version is not None:
                    self.base_sd_versions[hint] = self.current_sd_version

    def prepare_annotator(self, hint: ControlNetHints) -> None:
        if hint not in self.annotators:
            annotator_class = self.annotator_classes.get(hint)
            if annotator_class is None:
                raise ValueError(f"annotator for '{hint}' is not implemented")
            annotator = annotator_class(self.device)
            annotator = annotator.to(self.device, use_half=self.use_half)
            self.annotators[hint] = annotator

    def prepare_annotators(self) -> None:
        for hint in self.controlnet_weights:
            self.prepare_annotator(hint)

    def get_hint_of(
        self,
        hint: ControlNetHints,
        uint8_rgb: np.ndarray,
        **kwargs: Any,
    ) -> np.ndarray:
        if not self.loaded.get(hint):
            raise ValueError(f"'{hint}' is not loaded yet, please call `switch` first.")
        annotator = self.annotators.get(hint)
        if annotator is None:
            raise ValueError(
                f"annotator for '{hint}' is not prepared yet, "
                "please call `prepare_annotator`/`prepare_annotators` first."
            )
        kwargs["uint8_rgb"] = uint8_rgb
        out = safe_execute(annotator.annotate, kwargs)
        if len(out.shape) == 2:
            out = out[..., None]
        if out.shape[-1] == 1:
            out = np.repeat(out, 3, axis=2)
        return out

    def enable_control(self) -> None:
        self.m.control_model = self.control_model

    def disable_control(self) -> None:
        self.m.control_model = None


__all__ = [
    "SDVersions",
    "DiffusionAPI",
    "ControlNetHints",
    "ControlledDiffusionAPI",
]



WORKSPACE_ENVIRON_KEY = "CFLEARN_WORKSPACE"

LOSS_KEY = "loss"
INPUT_KEY = "input"
LATENT_KEY = "latent"
PREDICTIONS_KEY = "predictions"
LABEL_KEY = "labels"
ORIGINAL_LABEL_KEY = "original_labels"
BATCH_INDICES_KEY = "batch_indices"

PT_PREFIX = "model_"
SCORES_FILE = "scores.json"
CHECKPOINTS_FOLDER = "checkpoints"

DEFAULT_ZOO_TAG = "default"

# style transfer constants

STYLE_KEY = "style"
STYLE_LATENTS_KEY = "style_latents"
CONTENT_LATENT_KEY = "content_latent"
STYLIZED_STYLE_LATENTS_KEY = "stylized_style_latents"
STYLIZED_CONTENT_LATENT_KEY = "stylized_content_latent"

INPUT_B_KEY = "input_b"
LABEL_B_KEY = "labels_b"



from .blocks import *
from .ml import *
from .pytorch import *
from .array import *

# expose some other stuffs for convenience

from ..schema import DataConfig
from ..schema import DataProcessor
from ..schema import DataProcessorConfig



import torch

from torch import nn
from typing import Any
from typing import List
from typing import Optional
from cftool.misc import shallow_copy_dict

from ..register import register_backbone
from ......modules.blocks import ImgToPatches
from ......modules.blocks import MixedStackedEncoder


class MixViTBlock(nn.Module):
    def __init__(self, embed: ImgToPatches, encoder: MixedStackedEncoder):
        super().__init__()
        self.embed = embed
        self.encoder = encoder

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        net, hw = self.embed(net)
        net = self.encoder(net, hw)
        net = net.view(net.shape[0], *hw, -1).permute(0, 3, 1, 2).contiguous()
        return net


class MixViT(nn.Module):
    def __init__(
        self,
        in_channels: int,
        latent_dims: List[int],
        *,
        num_heads_list: List[int],
        feedforward_dim_ratios: List[float],
        num_layers_list: List[int],
        reduction_ratios: List[int],
        dropout: float = 0.0,
        drop_path_rate: float = 0.1,
        norm_type: Optional[str] = "layer",
        **attention_kwargs: Any,
    ):
        super().__init__()
        # patch embeddings
        patch_embeds = []
        patch_in_channels = in_channels
        for i, (patch_size, stride) in enumerate(zip([7, 3, 3, 3], [4, 2, 2, 2])):
            latent_dim = latent_dims[i]
            to_patches_config = {
                "img_size": 0,
                "patch_size": patch_size,
                "in_channels": patch_in_channels,
                "latent_dim": latent_dim,
                "stride": stride,
            }
            patch_embeds.append(ImgToPatches.make("overlap", to_patches_config))
            patch_in_channels = latent_dim
        # transformer encoders
        total_num_layers = sum(num_layers_list)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, total_num_layers)]
        cursor = 0
        encoders = []
        attention_kwargs.setdefault("bias", True)
        for i, (latent_dim, num_heads) in enumerate(zip(latent_dims, num_heads_list)):
            ff_ratio = feedforward_dim_ratios[i]
            num_layers = num_layers_list[i]
            ak = shallow_copy_dict(attention_kwargs)
            ak["num_heads"] = num_heads
            ak["reduction_ratio"] = reduction_ratios[i]
            encoders.append(
                MixedStackedEncoder(
                    latent_dim,
                    0,
                    token_mixing_type="attention",
                    token_mixing_config=ak,
                    channel_mixing_type="mix_ff",
                    num_layers=num_layers,
                    dropout=dropout,
                    dpr_list=dpr[cursor : cursor + num_layers],
                    norm_type=norm_type,
                    latent_dim_ratio=ff_ratio,
                    use_head_token=False,
                    head_pooler=None,
                    use_positional_encoding=False,
                )
            )
            cursor += num_layers
        # stages
        for i, (patch_embed, encoder) in enumerate(zip(patch_embeds, encoders)):
            setattr(self, f"stage{i + 1}", MixViTBlock(patch_embed, encoder))

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        net = self.stage1(net)
        net = self.stage2(net)
        net = self.stage3(net)
        net = self.stage4(net)
        return net


@register_backbone(
    "mix_vit",
    [64, 128, 320, 512],
    dict(
        stage1="stage1",
        stage2="stage2",
        stage3="stage3",
        stage4="stage4",
    ),
)
def mix_vit(pretrained: bool = False, *, in_channels: int = 3) -> MixViT:
    if pretrained:
        raise ValueError("`MixViT` does not support `pretrained`")
    return MixViT(
        in_channels,
        [64, 128, 320, 512],
        num_heads_list=[1, 2, 5, 8],
        feedforward_dim_ratios=[4.0] * 4,
        num_layers_list=[3, 4, 18, 3],
        reduction_ratios=[8, 4, 2, 1],
    )


@register_backbone(
    "mix_vit_lite",
    [32, 64, 160, 256],
    dict(
        stage1="stage1",
        stage2="stage2",
        stage3="stage3",
        stage4="stage4",
    ),
)
def mix_vit_lite(pretrained: bool = False, *, in_channels: int = 3) -> MixViT:
    if pretrained:
        raise ValueError("`MixViT` does not support `pretrained`")
    return MixViT(
        in_channels,
        [32, 64, 160, 256],
        num_heads_list=[1, 2, 5, 8],
        feedforward_dim_ratios=[4.0] * 4,
        num_layers_list=[2] * 4,
        reduction_ratios=[8, 4, 2, 1],
    )


@register_backbone(
    "mix_vit_large",
    [64, 128, 320, 512],
    dict(
        stage1="stage1",
        stage2="stage2",
        stage3="stage3",
        stage4="stage4",
    ),
)
def mix_vit_large(pretrained: bool = False, *, in_channels: int = 3) -> MixViT:
    if pretrained:
        raise ValueError("`MixViT` does not support `pretrained`")
    return MixViT(
        in_channels,
        [64, 128, 320, 512],
        num_heads_list=[1, 2, 5, 8],
        feedforward_dim_ratios=[4.0] * 4,
        num_layers_list=[3, 6, 40, 3],
        reduction_ratios=[8, 4, 2, 1],
    )


__all__ = [
    "mix_vit",
    "mix_vit_lite",
    "mix_vit_large",
]



import torch

from typing import Any
from typing import Dict
from typing import Optional
from cftool.types import tensor_dict_type

from .constants import MASK_KEY
from ....schema import TrainerState
from ....schema import IDLModel
from ....constants import INPUT_KEY
from ....constants import LATENT_KEY
from ....modules.blocks import MixedStackedEncoder


# TeT -> TextTransformer


@IDLModel.register("tet")
class TeTEncoder(IDLModel):
    def __init__(
        self,
        latent_dim: int = 384,
        context_length: int = 77,
        *,
        use_triu_attn_mask: bool = False,
        num_layers: int = 12,
        dropout: float = 0.0,
        drop_path_rate: float = 0.0,
        norm_position: str = "pre_norm",
        norm_type: Optional[str] = "layer",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        embedding_norm: Optional[torch.nn.Module] = None,
        embedding_dropout: Optional[float] = None,
        residual_after_norm: bool = False,
        feedforward_dim_ratio: float = 4.0,
        attention_kwargs: Optional[Dict[str, Any]] = None,
        feedforward_kwargs: Optional[Dict[str, Any]] = None,
        use_positional_encoding: bool = True,
        head_pooler: Optional[str] = None,
        no_head_norm: Optional[bool] = None,
        norm_after_head: bool = False,
    ):
        super().__init__()
        if not use_triu_attn_mask:
            self.attention_mask = None
        else:
            mask = torch.empty(context_length, context_length, dtype=torch.bool)
            mask.fill_(True)
            mask.triu_(1)
            self.register_buffer("attention_mask", mask)
        if attention_kwargs is None:
            attention_kwargs = {}
        attention_kwargs.setdefault("bias", True)
        attention_kwargs.setdefault("num_heads", 6)
        self.encoder = MixedStackedEncoder(
            latent_dim,
            context_length,
            token_mixing_type="attention",
            token_mixing_config=attention_kwargs,
            channel_mixing_config=feedforward_kwargs,
            num_layers=num_layers,
            dropout=dropout,
            drop_path_rate=drop_path_rate,
            norm_position=norm_position,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            embedding_norm=embedding_norm,
            embedding_dropout=embedding_dropout,
            residual_after_norm=residual_after_norm,
            latent_dim_ratio=feedforward_dim_ratio,
            use_head_token=False,
            head_pooler=head_pooler,
            use_positional_encoding=use_positional_encoding,
            is_vision_positional_encoding=False,
            no_head_norm=no_head_norm,
            norm_after_head=norm_after_head,
        )

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        *,
        apply_head: bool = True,
        clip_skip: int = 0,
        **kwargs: Any,
    ) -> tensor_dict_type:
        net = batch[INPUT_KEY]
        mask = batch.get(MASK_KEY, self.attention_mask)
        if mask is not None:
            t = net.shape[1]
            if t != mask.shape[0]:
                mask = mask[:t, :t]
        net = self.encoder.pre_process(net, **kwargs)
        for i, block in enumerate(self.encoder.mixing_blocks):
            net = block(net, mask=mask)
            if i == len(self.encoder.mixing_blocks) - 1 - clip_skip:
                break
        if apply_head:
            net = self.encoder.post_process(net)
        return {LATENT_KEY: net}


__all__ = ["TeTEncoder"]



import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

import torch
import numpy as np
from . import util
from .body import Body
from .hand import Hand
from .....misc.toolkit import download_model


class OpenposeDetector:
    def __init__(self, device: torch.device):
        body_modelpath = download_model("body_pose_model")
        hand_modelpath = download_model("hand_pose_model")

        self.device = device
        self.use_half = False
        self.body_estimation = Body(body_modelpath, device)
        self.hand_estimation = Hand(hand_modelpath, device)

    def to(self, device: torch.device, *, use_half: bool) -> None:
        if use_half:
            self.body_estimation.model.half()
            self.hand_estimation.model.half()
        else:
            self.body_estimation.model.float()
            self.hand_estimation.model.float()
        self.body_estimation.model.to(device)
        self.hand_estimation.model.to(device)
        self.device = device
        self.use_half = use_half

    def __call__(self, input_image, hand=False):
        def to(tensor):
            tensor = tensor.to(self.device)
            tensor = tensor.half() if self.use_half else tensor.float()
            return tensor

        input_image = input_image[..., ::-1].copy()
        with torch.no_grad():
            candidate, subset = self.body_estimation(input_image, to)
            canvas = np.zeros_like(input_image)
            canvas = util.draw_bodypose(canvas, candidate, subset)
            if hand:
                hands_list = util.handDetect(candidate, subset, input_image)
                all_hand_peaks = []
                for x, y, w, is_left in hands_list:
                    peaks = self.hand_estimation(
                        input_image[y : y + w, x : x + w, :], to
                    )
                    peaks[:, 0] = np.where(
                        peaks[:, 0] == 0, peaks[:, 0], peaks[:, 0] + x
                    )
                    peaks[:, 1] = np.where(
                        peaks[:, 1] == 0, peaks[:, 1], peaks[:, 1] + y
                    )
                    all_hand_peaks.append(peaks)
                canvas = util.draw_handpose(canvas, all_hand_peaks)
            return canvas, dict(candidate=candidate.tolist(), subset=subset.tolist())


__all__ = ["OpenposeDetector"]



import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Optional
from torch.nn import Module
from cftool.misc import shallow_copy_dict

from .basic import conv_nd
from .basic import get_conv_blocks
from ..utils import avg_pool_nd
from ..utils import zero_module
from ..utils import Residual
from ..hijacks import HijackConv2d
from ..hijacks import HijackLinear
from ....misc.toolkit import safe_clip_
from ....misc.toolkit import gradient_checkpoint


class ResidualBlock(Module):
    def __init__(
        self,
        dim: int,
        dropout: float,
        kernel_size: int = 3,
        stride: int = 1,
        *,
        ca_reduction: Optional[int] = None,
        eca_kernel_size: Optional[int] = None,
        norm_type: Optional[str] = "batch",
        **kwargs: Any,
    ):
        super().__init__()
        kwargs["norm_type"] = norm_type
        k1 = shallow_copy_dict(kwargs)
        k1["ca_reduction"] = ca_reduction
        k1.setdefault("activation", nn.LeakyReLU(0.2, inplace=True))
        blocks = get_conv_blocks(dim, dim, kernel_size, stride, **k1)
        if 0.0 < dropout < 1.0:
            blocks.append(nn.Dropout(dropout))
        k2 = shallow_copy_dict(kwargs)
        k2["activation"] = None
        k2["eca_kernel_size"] = eca_kernel_size
        blocks.extend(get_conv_blocks(dim, dim, kernel_size, stride, **k2))
        self.net = Residual(nn.Sequential(*blocks))

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


class ResidualBlockV2(Module):
    def __init__(
        self,
        dim: int,
        dropout: float,
        kernel_size: int = 3,
        stride: int = 1,
        *,
        ca_reduction: Optional[int] = None,
        eca_kernel_size: Optional[int] = None,
        norm_type: Optional[str] = "batch",
        **kwargs: Any,
    ):
        super().__init__()
        kwargs["norm_type"] = norm_type
        k1 = shallow_copy_dict(kwargs)
        k1["pre_activate"] = True
        k1["ca_reduction"] = ca_reduction
        k1.setdefault("activation", nn.LeakyReLU(0.2, inplace=True))
        blocks = get_conv_blocks(dim, dim, kernel_size, stride, **k1)
        if 0.0 < dropout < 1.0:
            blocks.append(nn.Dropout(dropout))
        k2 = shallow_copy_dict(kwargs)
        k2["pre_activate"] = True
        k2["eca_kernel_size"] = eca_kernel_size
        blocks.extend(get_conv_blocks(dim, dim, kernel_size, stride, **k2))
        self.net = Residual(nn.Sequential(*blocks))

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


class ResDownsample(Module):
    def __init__(
        self,
        in_channels: int,
        use_conv: bool,
        *,
        signal_dim: int = 2,
        out_channels: Optional[int] = None,
        padding: int = 1,
    ):
        super().__init__()
        out_channels = out_channels or in_channels
        stride = 2 if signal_dim != 3 else (1, 2, 2)
        if not use_conv:
            if in_channels != out_channels:
                raise ValueError(
                    "`in_channels` should be equal to `out_channels` "
                    "when `use_conv` is set to False"
                )
            self.net = avg_pool_nd(signal_dim, kernel_size=stride, stride=stride)
        else:
            self.net = conv_nd(
                signal_dim,
                in_channels,
                out_channels,
                3,
                stride=stride,
                padding=padding,
            )

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


class ResUpsample(Module):
    def __init__(
        self,
        in_channels: int,
        use_conv: bool,
        *,
        signal_dim: int = 2,
        out_channels: Optional[int] = None,
        padding: int = 1,
    ):
        super().__init__()
        self.signal_dim = signal_dim
        if not use_conv:
            self.conv = None
        else:
            self.conv = conv_nd(
                signal_dim,
                in_channels,
                out_channels,
                3,
                padding=padding,
            )

    def forward(self, net: Tensor) -> Tensor:
        if self.signal_dim == 3:
            _, _, c, h, w = net.shape
            net = F.interpolate(net, (c, h * 2, w * 2), mode="nearest")
        else:
            net = F.interpolate(net, scale_factor=2, mode="nearest")
        if self.conv is not None:
            net = self.conv(net)
        return net


class ResidualBlockWithTimeEmbedding(Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: Optional[int] = None,
        *,
        signal_dim: int = 2,
        dropout: float = 0.0,
        norm_eps: float = 1.0e-6,
        use_conv_shortcut: bool = False,
        integrate_upsample: bool = False,
        integrate_downsample: bool = False,
        time_embedding_channels: int = 512,
        use_scale_shift_norm: bool = False,
        use_checkpoint: bool = False,
    ):
        super().__init__()
        self.in_channels = in_channels
        out_channels = out_channels or in_channels
        self.out_channels = out_channels
        self.use_conv_shortcut = use_conv_shortcut
        self.use_scale_shift_norm = use_scale_shift_norm
        self.use_checkpoint = use_checkpoint

        self.resample = integrate_upsample or integrate_downsample
        if not self.resample:
            self.inp_resample = self.net_resample = None
        elif integrate_upsample:
            self.inp_resample = ResUpsample(in_channels, False, signal_dim=signal_dim)
            self.net_resample = ResUpsample(in_channels, False, signal_dim=signal_dim)
        else:
            self.inp_resample = ResDownsample(in_channels, False, signal_dim=signal_dim)
            self.net_resample = ResDownsample(in_channels, False, signal_dim=signal_dim)

        make_norm = lambda c: nn.GroupNorm(num_groups=32, num_channels=c, eps=norm_eps)

        self.activation = nn.SiLU()
        self.norm1 = make_norm(in_channels)
        self.conv1 = conv_nd(signal_dim, in_channels, out_channels, 3, 1, 1)
        if time_embedding_channels > 0:
            if use_scale_shift_norm:
                t_out_channels = 2 * out_channels
            else:
                t_out_channels = out_channels
            self.time_embedding = HijackLinear(time_embedding_channels, t_out_channels)
        self.norm2 = make_norm(out_channels)
        self.dropout = nn.Dropout(dropout)
        conv2 = conv_nd(signal_dim, out_channels, out_channels, 3, 1, 1)
        self.conv2 = zero_module(conv2)
        if in_channels != out_channels:
            if use_conv_shortcut:
                self.shortcut = HijackConv2d(in_channels, out_channels, 3, 1, 1)
            else:
                self.shortcut = HijackConv2d(in_channels, out_channels, 1, 1, 0)

    def forward(self, net: Tensor, time_net: Optional[Tensor] = None) -> Tensor:
        inputs = (net,) if time_net is None else (net, time_net)
        return gradient_checkpoint(
            self._forward,
            inputs=inputs,
            params=self.parameters(),
            enabled=self.use_checkpoint,
        )

    def _forward(self, net: Tensor, time_net: Optional[Tensor] = None) -> Tensor:
        inp = net
        net = self.norm1(net)
        net = self.activation(net)
        if self.inp_resample is None or self.net_resample is None:
            net = self.conv1(net)
        else:
            inp = self.inp_resample(inp)
            net = self.net_resample(net)
            net = self.conv1(net)
        if self.in_channels != self.out_channels:
            inp = self.shortcut(inp)

        if time_net is not None:
            time_net = self.activation(time_net)
            time_net = self.time_embedding(time_net)
            while len(time_net.shape) < len(net.shape):
                time_net = time_net[..., None]
            if self.use_scale_shift_norm:
                scale, shift = torch.chunk(time_net, 2, dim=1)
                net = self.norm2(net) * (1.0 + scale) + shift
                net = self.activation(net)
                net = self.dropout(net)
                net = self.conv2(net)
                return inp + net
            net = net + time_net

        net = self.norm2(net)
        net = self.activation(net)
        net = self.dropout(net)
        net = self.conv2(net)

        net = inp + net
        safe_clip_(net)

        return net


__all__ = [
    "ResidualBlock",
    "ResidualBlockV2",
    "ResidualBlockWithTimeEmbedding",
]



import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.misc import shallow_copy_dict
from cftool.array import iou
from cftool.array import corr
from cftool.array import to_torch
from cftool.types import tensor_dict_type

from ..types import losses_type
from ..schema import ILoss
from ..schema import TrainerState
from ..constants import LOSS_KEY
from ..constants import INPUT_KEY
from ..constants import LABEL_KEY
from ..constants import PREDICTIONS_KEY


@ILoss.register("iou")
class IOULoss(ILoss):
    def forward(self, logits: Tensor, labels: Tensor) -> Tensor:
        return 1.0 - iou(logits, labels)


@ILoss.register("bce")
class BCELoss(ILoss):
    def __init__(self, reduction: str = "mean"):
        super().__init__(reduction)
        self.bce = nn.BCEWithLogitsLoss(reduction="none")

    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        losses = self.bce(predictions, labels.to(predictions.dtype))
        return losses.mean(tuple(range(1, len(losses.shape))))


@ILoss.register("mae")
class MAELoss(ILoss):
    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        return F.l1_loss(predictions, labels, reduction="none")


@ILoss.register("sigmoid_mae")
class SigmoidMAELoss(ILoss):
    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        losses = F.l1_loss(torch.sigmoid(predictions), labels, reduction="none")
        return losses.mean((1, 2, 3))


@ILoss.register("mse")
class MSELoss(ILoss):
    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        return F.mse_loss(predictions, labels, reduction="none")


@ILoss.register("recon")
class ReconstructionLoss(ILoss):
    base_loss: ILoss

    def __init__(
        self, reduction: str = "mean", *, base_loss: str = "mae", **kwargs: Any
    ):
        super().__init__(reduction)
        self.base_loss = ILoss.make(base_loss, kwargs)

    def get_forward_args(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
    ) -> Tuple[Any, ...]:
        return forward_results, batch, state

    def forward(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
    ) -> losses_type:
        batch = shallow_copy_dict(batch)
        batch[LABEL_KEY] = batch[INPUT_KEY]
        return self.base_loss.run(forward_results, batch, state)


@ILoss.register("quantile")
class QuantileLoss(ILoss):
    def __init__(self, q: Union[float, List, Tuple, np.ndarray]):
        super().__init__()
        if isinstance(q, float):
            self.register_buffer("q", torch.tensor([q], torch.float32))
        else:
            q = np.asarray(q, np.float32).reshape([1, -1])
            self.register_buffer("q", to_torch(q))

    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        quantile_error = labels - predictions
        neg_errors = self.q * quantile_error  # type: ignore
        pos_errors = (self.q - 1.0) * quantile_error  # type: ignore
        quantile_losses = torch.max(neg_errors, pos_errors)
        return quantile_losses.mean(1, keepdim=True)


@ILoss.register("corr")
class CorrelationLoss(ILoss):
    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        return -corr(predictions, labels, get_diagonal=True)


@ILoss.register("cross_entropy")
class CrossEntropyLoss(ILoss):
    @staticmethod
    def _get_stat(predictions: Tensor, labels: Tensor) -> Tuple[Tensor, Tensor]:
        log_prob_mat = F.log_softmax(predictions, dim=1)
        nll_losses = -log_prob_mat.gather(dim=1, index=labels)
        return log_prob_mat, nll_losses

    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        return self._get_stat(predictions, labels)[1]


@ILoss.register("label_smooth_cross_entropy")
class LabelSmoothCrossEntropyLoss(ILoss):
    def __init__(self, eps: float = 0.1):
        super().__init__()
        self._eps = eps

    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        log_prob_mat, nll_losses = CrossEntropyLoss._get_stat(predictions, labels)
        smooth_losses = -log_prob_mat.sum(dim=1, keepdim=True)
        eps = self._eps / log_prob_mat.shape[-1]
        return (1.0 - self._eps) * nll_losses + eps * smooth_losses


@ILoss.register("focal")
class FocalLoss(ILoss):
    def __init__(
        self,
        *,
        input_logits: bool = True,
        eps: float = 1.0e-6,
        gamma: float = 2.0,
        alpha: Optional[Union[int, float, List, Tuple]] = None,
    ):
        super().__init__()
        self._input_logits = input_logits
        self._eps = eps
        self._gamma = gamma
        if isinstance(alpha, (int, float)):
            alpha = [alpha, 1 - alpha]
        elif isinstance(alpha, (list, tuple)):
            alpha = list(alpha)
        if alpha is None:
            self.alpha = None
        else:
            self.register_buffer("alpha", to_torch(np.array(alpha, np.float32)))

    def forward(self, predictions: Tensor, labels: Tensor) -> Tensor:
        if not self._input_logits:
            prob_mat = predictions.view(-1, predictions.shape[-1]) + self._eps
        else:
            logits_mat = predictions.view(-1, predictions.shape[-1])
            prob_mat = F.softmax(logits_mat, dim=1) + self._eps
        gathered_prob_flat = prob_mat.gather(dim=1, index=labels).view(-1)
        gathered_log_prob_flat = gathered_prob_flat.log()
        if self.alpha is not None:
            alpha_target = self.alpha.gather(dim=0, index=labels.view(-1))
            gathered_log_prob_flat = gathered_log_prob_flat * alpha_target
        loss = -gathered_log_prob_flat * (1 - gathered_prob_flat) ** self._gamma
        return loss.view_as(labels)


__all__ = [
    "IOULoss",
    "MAELoss",
    "MSELoss",
    "QuantileLoss",
    "SigmoidMAELoss",
    "CorrelationLoss",
    "CrossEntropyLoss",
    "LabelSmoothCrossEntropyLoss",
    "FocalLoss",
]



import torch

import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from itertools import combinations

from .base import MLModel
from ...schema import MLEncoderSettings
from ...schema import MLGlobalEncoderSettings
from ...misc.toolkit import eval_context
from ...modules.blocks import Activation


class SeparableMapping(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_groups: int,
        *,
        bias: bool = True,
        activation: Optional[str] = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.num_groups = num_groups
        in_dim *= num_groups
        out_dim *= num_groups
        self.conv = nn.Conv1d(in_dim, out_dim, 1, groups=num_groups, bias=bias)
        self.bn = None if not batch_norm else nn.BatchNorm1d(out_dim)
        self.activation = Activation.make(activation)
        use_dropout = 0.0 < dropout < 1.0
        self.dropout = None if not use_dropout else nn.Dropout(dropout)

    # [B, num_groups * in_dim, D] -> [B, num_groups * out_dim, D]
    # note that they should be originated from [B, num_groups, in_dim / out_dim, D]
    def forward(self, net: Tensor) -> Tensor:
        net = self.conv(net)
        if self.bn is not None:
            net = self.bn(net)
        net = self.activation(net)
        if self.dropout is not None:
            net = self.dropout(net)
        return net


class NBMBlock(nn.Module):
    def __init__(
        self,
        order: int,
        num_bases: int,
        *,
        hidden_units: List[int],
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        self.order = order
        self.num_bases = num_bases

        blocks: List[nn.Module] = []
        in_dim = order
        for hidden_unit in hidden_units:
            mapping = SeparableMapping(
                in_dim,
                hidden_unit,
                num_bases,
                bias=bias,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
            )
            blocks.append(mapping)
            in_dim = hidden_unit
        blocks.append(
            SeparableMapping(
                in_dim,
                1,
                num_bases,
                bias=bias,
                activation=None,
                batch_norm=False,
                dropout=0.0,
            )
        )

        self.net = nn.Sequential(*blocks)

    def forward(self, net: Tensor, indices: List[Tuple[int, ...]]) -> Tensor:
        for i, pack in enumerate(indices):
            if len(pack) != self.order:
                raise ValueError(
                    f"length of every tuple in indices must be {self.order}, "
                    f"but got {pack} at index {i}"
                )
        B = len(net)
        D = len(indices)
        # [B, D, order]
        net = net[..., indices]
        # [B, order, D]
        net = net.transpose(1, 2)
        # [B, 1, order, D]
        net = net.reshape(B, 1, self.order, D)
        # [B, num_bases, order, D]
        net = net.repeat_interleave(self.num_bases, 1)
        # [B, num_bases * order, D]
        net = net.contiguous().view(B, self.num_bases * self.order, D)
        # [B, num_bases, D]
        net = self.net(net)
        # [B, D, num_bases]
        return net.transpose(1, 2).contiguous()


@MLModel.register("nbm")
class NBM(MLModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int = 1,
        *,
        nary: Optional[Dict[str, List[Tuple[int, ...]]]] = None,
        num_bases: int = 64,
        bases_dropout: float = 0.0,
        hidden_units: Optional[List[int]] = None,
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
        encoder_settings: Optional[Dict[str, MLEncoderSettings]] = None,
        global_encoder_settings: Optional[MLGlobalEncoderSettings] = None,
    ):
        super().__init__(
            encoder_settings=encoder_settings,
            global_encoder_settings=global_encoder_settings,
        )
        if self.encoder is not None:
            input_dim += self.encoder.dim_increment
        input_dim *= num_history
        if hidden_units is None:
            dim = max(32, min(1024, 2 * input_dim))
            hidden_units = 2 * [dim]

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.bias = bias

        self.bases = nn.ModuleDict()
        self.nary = nary or {"1": list(combinations(range(input_dim), 1))}
        self.num_bases = num_bases
        for order in self.nary:
            self.bases[order] = NBMBlock(
                int(order),
                num_bases,
                hidden_units=hidden_units,
                bias=bias,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
            )
        self.bases_dropout = nn.Dropout(p=bases_dropout)

        self.all_indices = []
        for indices in self.nary.values():
            self.all_indices += indices
        self.latent_dim = len(self.all_indices)

        self.featurizer = nn.Conv1d(
            self.latent_dim * num_bases,
            self.latent_dim,
            kernel_size=1,
            groups=self.latent_dim,
            bias=bias,
        )
        self.head = nn.Linear(self.latent_dim, output_dim, bias=True)

    def get_key(self, order: str) -> str:
        return f"order_{order}"

    def forward(self, net: Tensor, *, return_features: bool = False) -> Tensor:
        B = len(net)
        if len(net.shape) > 2:
            net = net.contiguous().view(B, -1)
        bases = []
        for order, indices in self.nary.items():
            base = self.bases[order](net, indices)
            base = self.bases_dropout(base)
            bases.append(base)
        # [B, latent_dim, num_bases]
        net = torch.cat(bases, dim=1)
        # [B, latent_dim * num_bases, 1]
        net = net.view(B, self.latent_dim * self.num_bases, 1)
        # [B, latent_dim]
        net = self.featurizer(net).squeeze(2)
        if return_features:
            return net
        # [B, out_dim]
        net = self.head(net)
        return net

    def inspect(
        self,
        net: Tensor,
        x_dims: Tuple[int, ...],
        y_dim: int,
        *,
        already_extracted: bool = False,
    ) -> Tensor:
        try:
            idx = self.all_indices.index(x_dims)
        except ValueError:
            raise ValueError(f"`x_dims` ({x_dims}) is not in the pre-defined nary")
        with eval_context(self):
            B = len(net)
            if len(net.shape) > 2:
                net = net.contiguous().view(B, -1)
            order = len(x_dims)
            # [B, 1, num_bases]
            indices = [tuple(range(order))] if already_extracted else [x_dims]
            net = self.bases[str(order)](net, indices)
            net = self.bases_dropout(net)
            # [B, num_bases]
            net = net.view(B, self.num_bases, 1).squeeze(2)
            # [B, 1]
            w = self.featurizer.weight[[idx]].squeeze(2)
            b = None if not self.bias else self.featurizer.bias[[idx]]
            net = F.linear(net, w, b)
            # [B, 1]
            w = self.head.weight[..., [idx]][[y_dim]]
            net = F.linear(net, w)
        return net


__all__ = ["NBM"]



import math
import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .ddim import DDIMQSampler
from .utils import append_dims
from .utils import interpolate_fn
from .schema import ISampler
from .schema import IQSampler
from .schema import IDiffusion
from .schema import UncondSamplerMixin
from ..utils import cond_type
from .....misc.toolkit import get_device


@ISampler.register("solver")
class DPMSolver(ISampler, UncondSamplerMixin):
    def __init__(
        self,
        model: IDiffusion,
        *,
        schedule: str = "discrete",
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        t0: Optional[float] = None,
        tT: Optional[float] = None,
        order: int = 2,
        skip_type: str = "time_uniform",
        default_steps: int = 25,
        continuous_beta_0: float = 0.1,
        continuous_beta_1: float = 20.0,
        predict_x0: bool = True,
        thresholding: bool = False,
        threshold_max_val: float = 1.0,
    ):
        if model.parameterization != "eps":
            raise ValueError("only `eps` parameterization is supported in `DPMSolver`")
        super().__init__(model)
        if schedule not in ["discrete", "linear", "cosine"]:
            msg = "only (`discrete` | `linear` | `cosine`) can be used as `schedule`"
            raise ValueError(msg)
        # schedule
        if schedule == "discrete":
            default_tT = 1.0
            log_alphas = 0.5 * torch.log(model.alphas_cumprod)
            self.total_N = len(log_alphas)
            self.t_array = torch.linspace(0.0, 1.0, len(log_alphas) + 1)[1:].view(1, -1)
            self.log_alpha_array = log_alphas.view(1, -1)
        else:
            default_tT = 0.9946 if schedule == "cosine" else 1.0
            self.total_N = 1000
            self.beta_0 = continuous_beta_0
            self.beta_1 = continuous_beta_1
            self.cosine_s = 0.008
            self.cosine_beta_max = 999.0
            self.cosine_t_max = (
                math.atan(self.cosine_beta_max * (1.0 + self.cosine_s) / math.pi)
                * 2.0
                * (1.0 + self.cosine_s)
                / math.pi
                - self.cosine_s
            )
            self.cosine_log_alpha_0 = math.log(
                math.cos(self.cosine_s / (1.0 + self.cosine_s) * math.pi / 2.0)
            )
        # properties
        self.schedule = schedule
        self.unconditional_cond = unconditional_cond
        self.unconditional_guidance_scale = unconditional_guidance_scale
        self.t0 = 1.0 / self.total_N if t0 is None else t0
        self.tT = default_tT if tT is None else tT
        self.order = order
        self.skip_type = skip_type
        self.default_steps = default_steps
        self.predict_x0 = predict_x0
        self.thresholding = thresholding
        self.threshold_max_val = threshold_max_val

    @property
    def q_sampler(self) -> IQSampler:
        return self._q_sampler

    @property
    def sample_kwargs(self) -> Dict[str, Any]:
        return dict(
            unconditional_cond=self.unconditional_cond,
            unconditional_guidance_scale=self.unconditional_guidance_scale,
            t0=self.t0,
            tT=self.tT,
            order=self.order,
            skip_type=self.skip_type,
        )

    def sample_step(
        self,
        image: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
        *,
        unconditional_cond: Optional[Any] = None,
        unconditional_guidance_scale: float = 1.0,
        t0: float = 1.0e-4,
        tT: float = 1.0,
        order: int = 2,
        skip_type: str = "time_uniform",
        **kwargs: Any,
    ) -> Tensor:
        if step == 0:
            self._reset_buffers(
                total_step,
                unconditional_cond,
                unconditional_guidance_scale,
                t0,
                tT,
                skip_type,
            )
        b = image.shape[0]
        vec_t = self.timesteps[step].to(image).expand(b)
        if not self.t_prev_list:
            self.t_prev_list.append(vec_t)
            self.model_prev_list.append(
                self._model_fn(image, vec_t, cond, step, total_step)
            )
            return image
        if len(self.t_prev_list) < order:
            image = self._multistep_update(image, vec_t, len(self.t_prev_list))
            self.t_prev_list.append(vec_t)
            self.model_prev_list.append(
                self._model_fn(image, vec_t, cond, step, total_step)
            )
            return image
        image = self._multistep_update(image, vec_t, order)
        for i in range(order - 1):
            self.t_prev_list[i] = self.t_prev_list[i + 1]
            self.model_prev_list[i] = self.model_prev_list[i + 1]
        self.t_prev_list[-1] = vec_t
        if step < total_step - 1:
            self.model_prev_list[-1] = self._model_fn(
                image, vec_t, cond, step, total_step
            )
        # clear cache at last step
        else:
            self.t_prev_list.clear()
            self.model_prev_list.clear()
        return image

    # internal

    def _model_fn(
        self,
        x: Tensor,
        ts: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
    ) -> Tensor:
        if self.predict_x0:
            return self._image_fn(x, ts, cond, step, total_step)
        return self._noise_fn(x, ts, cond, step, total_step)

    def _noise_fn(
        self,
        x: Tensor,
        ts: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
    ) -> Tensor:
        ts = self.model.t * torch.max(ts - 1.0 / self.model.t, torch.zeros_like(ts))
        return self._uncond_denoise(x, ts, cond, step, total_step)

    def _image_fn(
        self,
        x: Tensor,
        ts: Tensor,
        cond: Optional[cond_type],
        step: int,
        total_step: int,
    ) -> Tensor:
        eps = self._noise_fn(x, ts, cond, step, total_step)
        ndim = x.dim()
        alpha_t = self._marginal_alpha(ts)
        sigma_t = self._marginal_std(ts)
        x0 = (x - append_dims(sigma_t, ndim) * eps) / append_dims(alpha_t, ndim)
        if self.thresholding:
            p = 0.995  # A hyperparameter in the paper of "Imagen" [1].
            s = torch.quantile(torch.abs(x0).reshape((x0.shape[0], -1)), p, dim=1)
            s = append_dims(
                torch.maximum(s, self.threshold_max_val * torch.ones_like(s)),
                ndim,
            )
            x0 = torch.clamp(x0, -s, s) / s
        return x0

    def _multistep_update(
        self,
        x: Tensor,
        vec_t: Tensor,
        order: int,
    ) -> Tensor:
        if order == 1:
            return self._first_update(
                x,
                self.t_prev_list[-1],
                vec_t,
                self.model_prev_list[-1],
            )
        if order == 2:
            return self._multistep_second_update(x, vec_t)
        if order == 3:
            return self._multistep_third_update(x, vec_t)
        raise ValueError(f"Solver order must be 1 or 2 or 3, got {order}")

    def _first_update(
        self,
        x: Tensor,
        vec_s: Tensor,
        vec_t: Tensor,
        model_s: Tensor,
    ) -> Tensor:
        ndim = x.dim()
        lambda_s = self._marginal_lambda(vec_s)
        lambda_t = self._marginal_lambda(vec_t)
        h = lambda_t - lambda_s
        log_alpha_s = self._marginal_log_mean_coef(vec_s)
        log_alpha_t = self._marginal_log_mean_coef(vec_t)
        sigma_s = self._marginal_std(vec_s)
        sigma_t = self._marginal_std(vec_t)
        if self.predict_x0:
            phi_1 = torch.expm1(-h)
            alpha_t = torch.exp(log_alpha_t)
            x_t = (
                append_dims(sigma_t / sigma_s, ndim) * x
                - append_dims(alpha_t * phi_1, ndim) * model_s
            )
            return x_t
        phi_1 = torch.expm1(h)
        x_t = (
            append_dims(torch.exp(log_alpha_t - log_alpha_s), ndim) * x
            - append_dims(sigma_t * phi_1, ndim) * model_s
        )
        return x_t

    def _multistep_second_update(self, x: Tensor, vec_t: Tensor) -> Tensor:
        ndim = x.dim()
        t_prev_1, t_prev_0 = self.t_prev_list
        model_prev_1, model_prev_0 = self.model_prev_list
        lambda_prev_1 = self._marginal_lambda(t_prev_1)
        lambda_prev_0 = self._marginal_lambda(t_prev_0)
        lambda_t = self._marginal_lambda(vec_t)
        log_alpha_t = self._marginal_log_mean_coef(vec_t)
        sigma_t = self._marginal_std(vec_t)

        h_0 = lambda_prev_0 - lambda_prev_1
        h = lambda_t - lambda_prev_0
        r0 = h_0 / h
        d1_0 = append_dims(1.0 / r0, ndim) * (model_prev_0 - model_prev_1)

        if self.predict_x0:
            exp_mh_m1 = torch.exp(-h) - 1.0
            sigma_prev_0 = self._marginal_std(t_prev_0)
            alpha_t = torch.exp(log_alpha_t)
            x_t = (
                append_dims(sigma_t / sigma_prev_0, ndim) * x
                - append_dims(alpha_t * exp_mh_m1, ndim) * model_prev_0
                - 0.5 * append_dims(alpha_t * exp_mh_m1, ndim) * d1_0
            )
            return x_t
        exp_h_m1 = torch.exp(h) - 1.0
        log_alpha_prev_0 = self._marginal_log_mean_coef(t_prev_0)
        x_t = (
            append_dims(torch.exp(log_alpha_t - log_alpha_prev_0), ndim) * x
            - append_dims(sigma_t * exp_h_m1, ndim) * model_prev_0
            - 0.5 * append_dims(sigma_t * exp_h_m1, ndim) * d1_0
        )
        return x_t

    def _multistep_third_update(self, x: Tensor, vec_t: Tensor) -> Tensor:
        ndim = x.dim()
        t_prev_2, t_prev_1, t_prev_0 = self.t_prev_list
        model_prev_2, model_prev_1, model_prev_0 = self.model_prev_list
        lambda_prev_2 = self._marginal_lambda(t_prev_2)
        lambda_prev_1 = self._marginal_lambda(t_prev_1)
        lambda_prev_0 = self._marginal_lambda(t_prev_0)
        lambda_t = self._marginal_lambda(vec_t)
        log_alpha_t = self._marginal_log_mean_coef(vec_t)
        sigma_t = self._marginal_std(vec_t)

        h_1 = lambda_prev_1 - lambda_prev_2
        h_0 = lambda_prev_0 - lambda_prev_1
        h = lambda_t - lambda_prev_0
        r0, r1 = h_0 / h, h_1 / h
        d1_0 = append_dims(1.0 / r0, ndim) * (model_prev_0 - model_prev_1)
        d1_1 = append_dims(1.0 / r1, ndim) * (model_prev_1 - model_prev_2)
        d1 = d1_0 + append_dims(r0 / (r0 + r1), ndim) * (d1_0 - d1_1)
        d2 = append_dims(1.0 / (r0 + r1), ndim) * (d1_0 - d1_1)

        if self.predict_x0:
            exp_mh_m1 = torch.exp(-h) - 1.0
            sigma_prev_0 = self._marginal_std(t_prev_0)
            alpha_t = torch.exp(log_alpha_t)
            x_t = (
                append_dims(sigma_t / sigma_prev_0, ndim) * x
                - append_dims(alpha_t * exp_mh_m1, ndim) * model_prev_0
                + append_dims(alpha_t * (exp_mh_m1 / h + 1.0), ndim) * d1
                - append_dims(alpha_t * ((exp_mh_m1 + h) / h**2 - 0.5), ndim) * d2
            )
            return x_t
        exp_h_m1 = torch.exp(h) - 1.0
        log_alpha_prev_0 = self._marginal_log_mean_coef(t_prev_0)
        x_t = (
            append_dims(torch.exp(log_alpha_t - log_alpha_prev_0), ndim) * x
            - append_dims(sigma_t * exp_h_m1, ndim) * model_prev_0
            - append_dims(sigma_t * (exp_h_m1 / h - 1.0), ndim) * d1
            - append_dims(sigma_t * ((exp_h_m1 - h) / h**2 - 0.5), ndim) * d2
        )
        return x_t

    def _reset_buffers(
        self,
        total_step: int,
        unconditional_cond: Optional[Any],
        unconditional_guidance_scale: float,
        t0: float,
        tT: float,
        skip_type: str,
    ) -> None:
        self._q_sampler = DDIMQSampler(self.model)
        self._q_sampler.reset_buffers("uniform", total_step)
        self.timesteps = self._get_time_steps(skip_type, t0, tT, total_step - 1)
        assert self.timesteps.shape[0] == total_step
        self.t_prev_list: List[Tensor] = []
        self.model_prev_list: List[Tensor] = []
        # unconditional conditioning
        self._reset_uncond_buffers(unconditional_cond, unconditional_guidance_scale)

    def _get_time_steps(self, skip_type: str, t0: float, tT: float, N: int) -> Tensor:
        device = get_device(self.model)
        if skip_type == "logSNR":
            lambda_T = self._marginal_lambda(torch.tensor(tT, device=device))
            lambda_0 = self._marginal_lambda(torch.tensor(t0, device=device))
            logSNR_steps = torch.linspace(lambda_T, lambda_0, N + 1, device=device)
            return self._inverse_lambda(logSNR_steps)
        if skip_type == "time_uniform":
            return torch.linspace(tT, t0, N + 1, device=device)
        if skip_type == "time_quadratic":
            t = torch.linspace(t0, tT, 10000000, device=device)
            quadratic_t = torch.sqrt(t)
            quadratic_steps = torch.linspace(quadratic_t[0], quadratic_t[-1], N + 1)
            return torch.flip(
                torch.cat(
                    [
                        t[torch.searchsorted(quadratic_t, quadratic_steps)[:-1]],
                        tT * torch.ones((1,)).to(device),
                    ],
                    dim=0,
                ),
                dims=[0],
            )
        raise ValueError(f"unrecognized skip_type '{skip_type}' occurred")

    def _marginal_log_mean_coef(self, t: Tensor) -> Tensor:
        if self.schedule == "discrete":
            t = t.view(-1, 1)
            self.t_array = self.t_array.to(t)
            self.log_alpha_array = self.log_alpha_array.to(t)
            return interpolate_fn(t, self.t_array, self.log_alpha_array).view(-1)
        if self.schedule == "linear":
            return -0.25 * t**2 * (self.beta_1 - self.beta_0) - 0.5 * t * self.beta_0
        if self.schedule == "cosine":
            log_alpha_t = torch.log(
                torch.cos((t + self.cosine_s) / (1.0 + self.cosine_s) * math.pi / 2.0)
            )
            log_alpha_t = log_alpha_t - self.cosine_log_alpha_0
            return log_alpha_t
        raise ValueError(f"unrecognized schedule '{self.schedule}' occurred")

    def _marginal_alpha(self, t: Tensor) -> Tensor:
        return torch.exp(self._marginal_log_mean_coef(t))

    def _marginal_std(self, t: Tensor) -> Tensor:
        return torch.sqrt(1.0 - torch.exp(2.0 * self._marginal_log_mean_coef(t)))

    def _marginal_lambda(self, t: Tensor) -> Tensor:
        log_mean_coef = self._marginal_log_mean_coef(t)
        log_std = 0.5 * torch.log(1.0 - torch.exp(2.0 * log_mean_coef))
        return log_mean_coef - log_std

    def _inverse_lambda(self, lamb: Tensor) -> Tensor:
        if self.schedule == "discrete":
            self.t_array = self.t_array.to(lamb)
            self.log_alpha_array = self.log_alpha_array.to(lamb)
            zeros = torch.zeros((1,), dtype=lamb.dtype, device=lamb.device)
            log_alpha = (-0.5 * torch.logaddexp(zeros, -2.0 * lamb)).view(-1, 1)
            t = interpolate_fn(
                log_alpha,
                torch.flip(self.log_alpha_array, [1]),
                torch.flip(self.t_array, [1]),
            )
            return t.view(-1)
        if self.schedule == "linear":
            tmp = (
                2.0
                * (self.beta_1 - self.beta_0)
                * torch.logaddexp(-2.0 * lamb, torch.zeros((1,)).to(lamb))
            )
            delta = self.beta_0**2 + tmp
            return tmp / (torch.sqrt(delta) + self.beta_0) / (self.beta_1 - self.beta_0)
        if self.schedule == "cosine":
            log_alpha = -0.5 * torch.logaddexp(-2.0 * lamb, torch.zeros((1,)).to(lamb))
            t = (
                torch.arccos(torch.exp(log_alpha + self.cosine_log_alpha_0))
                * 2.0
                * (1.0 + self.cosine_s)
                / math.pi
                - self.cosine_s
            )
            return t
        raise ValueError(f"unrecognized schedule '{self.schedule}' occurred")


__all__ = [
    "DPMSolver",
]



import cflearn

from ._utils import get_info


if __name__ == "__main__":
    info = get_info()
    data = info.data
    assert data is not None
    cuda = info.meta["cuda"]
    assert info.config is not None
    cflearn.MLTrainingPipeline.init(info.config).fit(data, cuda=cuda)



import torch

from torch import Tensor
from typing import Union
from cftool.array import tensor_dict_type


cond_type = Union[Tensor, tensor_dict_type]
ADM_KEY = "labels"
ADM_TYPE = "adm"
CONCAT_KEY = "concat"
CONCAT_TYPE = "concat"
HYBRID_TYPE = "hybrid"
CROSS_ATTN_KEY = "context"
CROSS_ATTN_TYPE = "cross_attn"
CONTROL_HINT_KEY = "hint"
CONTROL_HINT_START_KEY = "hint_start"


def extract_to(array: Tensor, indices: Tensor, num_dim: int) -> Tensor:
    b = indices.shape[0]
    out = array.gather(-1, indices).contiguous()
    return out.view(b, *([1] * (num_dim - 1)))


def get_timesteps(t: int, num: int, device: torch.device) -> Tensor:
    return torch.full((num,), t, device=device, dtype=torch.long)



from typing import Callable

from .schema import IMetric
from .misc.toolkit import Initializer


def register_initializer(name: str) -> Callable:
    def _register(f: Callable) -> Callable:
        Initializer.add_initializer(f, name)
        return f

    return _register


def register_metric(name: str, *, allow_duplicate: bool = False) -> Callable:
    return IMetric.register(name, allow_duplicate=allow_duplicate)


__all__ = [
    "register_initializer",
    "register_metric",
]



import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Tuple
from typing import Optional

from .schema import DecoderMixin
from ....modules.blocks import make_attention
from ....modules.blocks import HijackConv2d
from ....modules.blocks import ApplyTanhMixin
from ....modules.blocks import ResidualBlockWithTimeEmbedding


class Upsample(nn.Module):
    def __init__(self, in_channels: int, with_conv: bool):
        super().__init__()
        if not with_conv:
            self.conv = None
        else:
            self.conv = HijackConv2d(in_channels, in_channels, 3, 1, 1)

    def forward(self, net: Tensor) -> Tensor:
        net = F.interpolate(net, scale_factor=2.0, mode="nearest")
        if self.conv is not None:
            net = self.conv(net)
        return net


@DecoderMixin.register("attention")
class AttentionDecoder(nn.Module, DecoderMixin, ApplyTanhMixin):
    def __init__(
        self,
        img_size: int,
        out_channels: int,
        inner_channels: int,
        latent_channels: int,
        channel_multipliers: Tuple[int, ...],
        *,
        num_res_blocks: int,
        attention_resolutions: Tuple[int, ...] = (),
        dropout: float = 0.0,
        upsample_with_conv: bool = True,
        attention_type: str = "spatial",
        apply_tanh: bool = False,
    ):
        super().__init__()
        self.img_size = img_size
        self.out_channels = out_channels
        self.inner_channels = inner_channels
        self.latent_channels = latent_channels
        self.num_upsample = len(channel_multipliers)
        self.num_res_blocks = num_res_blocks
        self.apply_tanh = apply_tanh
        make_res_block = lambda in_c, out_c: ResidualBlockWithTimeEmbedding(
            in_c,
            out_c,
            dropout=dropout,
            time_embedding_channels=0,
        )
        # in conv
        in_nc = inner_channels * channel_multipliers[-1]
        blocks = [HijackConv2d(latent_channels, in_nc, 3, 1, 1)]
        # residual
        blocks += [
            make_res_block(in_nc, in_nc),
            make_attention(in_nc, attention_type),
            make_res_block(in_nc, in_nc),
        ]
        # upsample
        current_resolution = img_size // 2 ** (self.num_upsample - 1)
        for i in reversed(range(self.num_upsample)):
            out_nc = inner_channels * channel_multipliers[i]
            for _ in range(self.num_res_blocks + 1):
                blocks.append(make_res_block(in_nc, out_nc))
                in_nc = out_nc
                if current_resolution in attention_resolutions:
                    blocks.append(make_attention(in_nc, attention_type))
            if i != 0:
                blocks.append(Upsample(in_nc, upsample_with_conv))
                current_resolution *= 2
        # head
        head_blocks = [
            nn.GroupNorm(num_groups=32, num_channels=in_nc, eps=1.0e-6, affine=True),
            nn.SiLU(),
            HijackConv2d(in_nc, out_channels, 3, 1, 1),
        ]
        # construct
        self.decoder = nn.Sequential(*blocks)
        self.head = nn.Sequential(*head_blocks)

    def forward(
        self,
        net: Tensor,
        *,
        no_head: bool = False,
        apply_tanh: Optional[bool] = None,
    ) -> Tensor:
        net = self.decoder(net)
        if no_head:
            return net
        net = self.head(net)
        net = self.postprocess(net, apply_tanh)
        return net


__all__ = [
    "AttentionDecoder",
]



import torch.nn as nn

from .ih_model import IHModelWithBackbone
from ..modifiers import LRMult
from ..modeling.deeplab_v3 import DeepLabV3Plus
from ..modeling.basic_blocks import MaxPoolDownSize


class DeepLabIHModel(IHModelWithBackbone):
    def __init__(
        self,
        base_config,
        mask_fusion="sum",
        deeplab_backbone="resnet34",
        lr_mult=0.1,
        pyramid_channels=-1,
        deeplab_ch=256,
        mode="cat",
        **base_kwargs
    ):
        """
        Creates image harmonization model supported by the features extracted from the pre-trained DeepLab backbone.

        Parameters
        ----------
        base_config : dict
            Configuration dict for the base model, to which the backbone features are incorporated.
            base_config contains model class and init parameters, examples can be found in iharm.mconfigs.base_models
        mask_fusion : str
            How to fuse the binary mask with the backbone input:
            'sum': apply convolution to the mask and sum it with the output of the first convolution in the backbone
            'rgb': concatenate the mask to the input image and translate it back to 3 channels with convolution
            otherwise: do not fuse mask with the backbone input
        deeplab_backbone : str
            ResNet backbone name.
        lr_mult : float
            Multiply learning rate to lr_mult when updating the weights of the backbone.
        pyramid_channels : int
            The DeepLab output can be consequently downsized to produce a feature pyramid.
            The pyramid features are then fused with the encoder outputs in the base model on multiple layers.
            Each pyramid feature map contains equal number of channels equal to pyramid_channels.
            If pyramid_channels <= 0, the feature pyramid is not constructed.
        deeplab_ch : int
            Number of channels for output DeepLab layer and some in the middle.
        mode : str
            How to fuse the backbone features with the encoder outputs in the base model:
            'sum': apply convolution to the backbone feature map obtaining number of channels
             same as in the encoder output and sum them
            'cat': concatenate the backbone feature map with the encoder output
            'catc': concatenate the backbone feature map with the encoder output and apply convolution obtaining
            number of channels same as in the encoder output
            otherwise: the backbone features are not incorporated into the base model
        base_kwargs : dict
            any kwargs associated with the base model
        """
        params = base_config["params"]
        params.update(base_kwargs)
        depth = params["depth"]

        backbone = DeepLabBB(pyramid_channels, deeplab_ch, deeplab_backbone, lr_mult)

        downsize_input = depth > 7
        params.update(
            dict(
                backbone_from=3 if downsize_input else 2,
                backbone_channels=backbone.output_channels,
                backbone_mode=mode,
            )
        )
        base_model = base_config["model"](**params)

        super(DeepLabIHModel, self).__init__(
            base_model, backbone, downsize_input, mask_fusion
        )


class DeepLabBB(nn.Module):
    def __init__(
        self,
        pyramid_channels=256,
        deeplab_ch=256,
        backbone="resnet34",
        backbone_lr_mult=0.1,
    ):
        super(DeepLabBB, self).__init__()
        self.pyramid_on = pyramid_channels > 0
        if self.pyramid_on:
            self.output_channels = [pyramid_channels] * 4
        else:
            self.output_channels = [deeplab_ch]

        self.deeplab = DeepLabV3Plus(
            backbone=backbone,
            ch=deeplab_ch,
            project_dropout=0.2,
            norm_layer=nn.BatchNorm2d,
            backbone_norm_layer=nn.BatchNorm2d,
        )
        self.deeplab.backbone.apply(LRMult(backbone_lr_mult))

        if self.pyramid_on:
            self.downsize = MaxPoolDownSize(
                deeplab_ch, pyramid_channels, pyramid_channels, 4
            )

    def forward(self, image, mask, mask_features):
        outputs = list(self.deeplab(image, mask_features))
        if self.pyramid_on:
            outputs = self.downsize(outputs[0])
        return outputs

    def load_pretrained_weights(self):
        self.deeplab.load_pretrained_weights()



import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Optional
from functools import partial

from .schema import ISpecializedConditionModel
from .....modules.blocks import HijackConv2d


@ISpecializedConditionModel.register("rescaler")
class SpatialRescaler(ISpecializedConditionModel):
    def __init__(
        self,
        *,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        num_stages: int = 1,
        multiplier: float = 0.5,
        method: str = "bilinear",
        bias: bool = False,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.num_stages = num_stages
        self.factor = round(1.0 / multiplier**num_stages)
        supported = {"nearest", "linear", "bilinear", "trilinear", "bicubic", "area"}
        if method not in supported:
            raise ValueError(f"`method` should be one of {supported}")
        self.interpolator = partial(F.interpolate, mode=method, scale_factor=multiplier)
        if out_channels is None:
            self.channel_mapper = None
        else:
            self.channel_mapper = HijackConv2d(in_channels, out_channels, 1, bias=bias)

    def forward(self, cond: Tensor) -> Tensor:
        for _ in range(self.num_stages):
            cond = self.interpolator(cond)
        if self.channel_mapper is not None:
            cond = self.channel_mapper(cond)
        return cond


__all__ = [
    "SpatialRescaler",
]



import math
import torch

from abc import abstractmethod
from abc import ABCMeta
from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional

from ....modules.blocks import conv_nd
from ....modules.blocks import zero_module
from ....modules.blocks import HijackLinear
from ....modules.blocks import SpatialTransformer
from ....modules.blocks import MultiHeadSpatialAttention
from ....modules.blocks import ResidualBlockWithTimeEmbedding
from ....modules.blocks.convs.residual import ResUpsample
from ....modules.blocks.convs.residual import ResDownsample


class TimestepBlock(nn.Module, metaclass=ABCMeta):
    @abstractmethod
    def forward(self, net: Tensor, time_net: Tensor) -> Tensor:
        pass


class TimestepAttnSequential(TimestepBlock, nn.Sequential, metaclass=ABCMeta):
    def forward(
        self,
        net: Tensor,
        time_net: Tensor,
        context: Optional[Tensor] = None,
    ) -> Tensor:
        for layer in self:
            if isinstance(layer, TimestepBlock):
                net = layer(net, time_net)
            elif isinstance(layer, SpatialTransformer):
                net = layer(net, context)
            else:
                net = layer(net)
        return net


class ResBlock(ResidualBlockWithTimeEmbedding, TimestepBlock):
    pass


def timestep_embedding(
    # 1-D Tensor, shape=[B]
    timesteps: Tensor,
    output_dim: int,
    *,
    dtype: torch.dtype,
    max_period: int = 10000,
    repeat_only: bool = False,
) -> Tensor:
    if repeat_only:
        return timesteps[..., None].repeat_interleave(output_dim, dim=1)
    half = output_dim // 2
    frequency = torch.exp(
        -math.log(max_period)
        * torch.arange(start=0, end=half, dtype=torch.float32)
        / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * frequency[None]
    embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
    if output_dim % 2:
        embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    embedding = embedding.to(dtype)
    return embedding


class UNetDiffuser(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        num_heads: Optional[int] = None,
        num_head_channels: Optional[int] = None,
        use_spatial_transformer: bool = False,
        num_transformer_layers: int = 1,
        context_dim: Optional[int] = None,
        signal_dim: int = 2,
        start_channels: int = 320,
        num_res_blocks: int = 2,
        attention_downsample_rates: Tuple[int, ...] = (1, 2, 4),
        dropout: float = 0.0,
        channel_multipliers: Tuple[int, ...] = (1, 2, 4, 8),
        resample_with_conv: bool = True,
        resample_with_resblock: bool = False,
        use_scale_shift_norm: bool = False,
        num_classes: Optional[int] = None,
        use_linear_in_transformer: bool = False,
        # misc
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        tome_info: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.context_dim = context_dim
        self.num_heads = num_heads
        self.num_head_channels = num_head_channels
        self.signal_dim = signal_dim
        self.start_channels = start_channels
        self.num_res_blocks = num_res_blocks
        self.attention_downsample_rates = attention_downsample_rates
        self.dropout = dropout
        self.channel_multipliers = channel_multipliers
        self.resample_with_conv = resample_with_conv
        self.use_scale_shift_norm = use_scale_shift_norm
        self.num_classes = num_classes
        self.use_linear_in_transformer = use_linear_in_transformer
        self.use_checkpoint = use_checkpoint
        self.attn_split_chunk = attn_split_chunk

        time_embedding_dim = start_channels * 4
        self.time_embedding = nn.Sequential(
            HijackLinear(start_channels, time_embedding_dim),
            nn.SiLU(),
            HijackLinear(time_embedding_dim, time_embedding_dim),
        )

        if num_classes is None:
            self.label_embedding = None
        else:
            self.label_embedding = nn.Embedding(num_classes, time_embedding_dim)

        make_res_block = lambda in_c, out_c, **kwargs: ResBlock(
            in_c,
            out_c,
            signal_dim=signal_dim,
            norm_eps=1.0e-5,
            time_embedding_channels=time_embedding_dim,
            dropout=dropout,
            use_checkpoint=use_checkpoint,
            use_scale_shift_norm=use_scale_shift_norm,
            **kwargs,
        )

        def make_attn_block(in_c: int) -> nn.Module:
            err_msg = "either `num_heads` or `num_head_channels` should be provided"
            if num_head_channels is not None:
                n_heads = in_c // num_head_channels
            else:
                if num_heads is None:
                    raise ValueError(err_msg)
                n_heads = num_heads
            head_c = in_c // n_heads if num_head_channels is None else num_head_channels
            if not use_spatial_transformer:
                # TODO: support attn_split_chunk
                return MultiHeadSpatialAttention(
                    in_c,
                    num_heads=n_heads,
                    num_head_channels=head_c,
                    use_checkpoint=use_checkpoint,
                )
            return SpatialTransformer(
                in_c,
                n_heads,
                head_c,
                num_layers=num_transformer_layers,
                context_dim=context_dim,
                use_linear=use_linear_in_transformer,
                use_checkpoint=use_checkpoint,
                attn_split_chunk=attn_split_chunk,
                tome_info=tome_info,
            )

        def make_downsample(in_c: int, out_c: int) -> TimestepAttnSequential:
            if not resample_with_resblock:
                return TimestepAttnSequential(
                    ResDownsample(
                        in_c,
                        resample_with_conv,
                        signal_dim=signal_dim,
                        out_channels=out_c,
                    )
                )
            res_block = make_res_block(in_c, out_c, integrate_downsample=True)
            return TimestepAttnSequential(res_block)

        def make_upsample(in_c: int, out_c: int) -> nn.Module:
            if not resample_with_resblock:
                return ResUpsample(
                    in_c,
                    resample_with_conv,
                    signal_dim=signal_dim,
                    out_channels=out_c,
                )
            return make_res_block(in_c, out_c, integrate_upsample=True)

        # input
        input_blocks = [
            TimestepAttnSequential(
                conv_nd(signal_dim, in_channels, start_channels, 3, padding=1)
            ),
        ]
        feature_size = start_channels
        input_block_channels = [start_channels]
        in_nc = start_channels
        downsample_rate = 1
        for i, multiplier in enumerate(channel_multipliers):
            for _ in range(num_res_blocks):
                out_nc = multiplier * start_channels
                blocks = [make_res_block(in_nc, out_nc)]
                in_nc = out_nc
                if downsample_rate in attention_downsample_rates:
                    blocks.append(make_attn_block(in_nc))
                input_blocks.append(TimestepAttnSequential(*blocks))
                feature_size += in_nc
                input_block_channels.append(in_nc)
            if i != len(channel_multipliers) - 1:
                out_nc = in_nc
                input_blocks.append(make_downsample(in_nc, out_nc))
                downsample_rate *= 2
                feature_size += in_nc
                input_block_channels.append(in_nc)
        self.input_blocks = nn.ModuleList(input_blocks)

        # residual
        self.residual = TimestepAttnSequential(
            make_res_block(in_nc, in_nc),
            make_attn_block(in_nc),
            make_res_block(in_nc, in_nc),
        )
        feature_size += in_nc

        # output
        output_blocks = []
        for i, multiplier in list(enumerate(channel_multipliers))[::-1]:
            for idx in range(num_res_blocks + 1):
                idx_nc = input_block_channels.pop()
                out_nc = start_channels * multiplier
                blocks = [make_res_block(in_nc + idx_nc, out_nc)]
                in_nc = out_nc
                if downsample_rate in attention_downsample_rates:
                    blocks.append(make_attn_block(in_nc))
                if i != 0 and idx == num_res_blocks:
                    out_nc = in_nc
                    blocks.append(make_upsample(in_nc, out_nc))
                    downsample_rate //= 2
                output_blocks.append(TimestepAttnSequential(*blocks))
                feature_size += in_nc
        self.output_blocks = nn.ModuleList(output_blocks)

        # head
        head_conv = conv_nd(signal_dim, start_channels, out_channels, 3, padding=1)
        self.head = nn.Sequential(
            nn.GroupNorm(32, in_nc),
            nn.SiLU(),
            zero_module(head_conv),
        )

    def set_tome_info(self, tome_info: Optional[Dict[str, Any]]) -> None:
        for m in self.modules():
            if isinstance(m, SpatialTransformer):
                m.set_tome_info(tome_info)

    def forward(
        self,
        net: Tensor,
        *,
        timesteps: Tensor,
        context: Optional[Tensor] = None,
        labels: Optional[Tensor] = None,
        control: Optional[List[Tensor]] = None,
        only_mid_control: bool = False,
    ) -> Tensor:
        if (labels is None) ^ (self.num_classes is None):
            raise ValueError("`labels` should be given iff `num_classes` is specified")

        # tomesd
        for m in self.modules():
            if isinstance(m, SpatialTransformer):
                for block in m.blocks:
                    if block.tome_info is not None:
                        block.tome_info["size"] = net.shape[-2:]

        # timenet
        time_net = timestep_embedding(
            timesteps,
            self.start_channels,
            dtype=net.dtype,
            repeat_only=False,
        )
        time_net = self.time_embedding(time_net)
        if self.label_embedding is not None:
            time_net = time_net + self.label_embedding(labels)

        # main
        nets = []
        for block in self.input_blocks:
            net = block(net, time_net, context)
            nets.append(net)
        net = self.residual(net, time_net, context)
        if control is not None:
            net += control.pop()
        for block in self.output_blocks:
            if only_mid_control or control is None:
                net = torch.cat([net, nets.pop()], dim=1)
            else:
                net = torch.cat([net, nets.pop() + control.pop()], dim=1)
            net = block(net, time_net, context)
        net = self.head(net)

        return net


class ControlNet(nn.Module):
    def __init__(
        self,
        in_channels: int,
        hint_channels: int,
        *,
        num_heads: Optional[int] = None,
        num_head_channels: Optional[int] = None,
        use_spatial_transformer: bool = False,
        num_transformer_layers: int = 1,
        context_dim: Optional[int] = None,
        signal_dim: int = 2,
        start_channels: int = 320,
        num_res_blocks: int = 2,
        attention_downsample_rates: Tuple[int, ...] = (1, 2, 4),
        dropout: float = 0.0,
        channel_multipliers: Tuple[int, ...] = (1, 2, 4, 8),
        resample_with_conv: bool = True,
        resample_with_resblock: bool = False,
        use_scale_shift_norm: bool = False,
        num_classes: Optional[int] = None,
        use_linear_in_transformer: bool = False,
        # misc
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        tome_info: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.hint_channels = hint_channels
        self.context_dim = context_dim
        self.num_heads = num_heads
        self.num_head_channels = num_head_channels
        self.signal_dim = signal_dim
        self.start_channels = start_channels
        self.num_res_blocks = num_res_blocks
        self.attention_downsample_rates = attention_downsample_rates
        self.dropout = dropout
        self.channel_multipliers = channel_multipliers
        self.resample_with_conv = resample_with_conv
        self.use_scale_shift_norm = use_scale_shift_norm
        self.num_classes = num_classes
        self.use_linear_in_transformer = use_linear_in_transformer
        self.use_checkpoint = use_checkpoint
        self.attn_split_chunk = attn_split_chunk

        time_embedding_dim = start_channels * 4
        self.time_embed = nn.Sequential(
            HijackLinear(start_channels, time_embedding_dim),
            nn.SiLU(),
            HijackLinear(time_embedding_dim, time_embedding_dim),
        )

        if num_classes is None:
            self.label_embedding = None
        else:
            self.label_embedding = nn.Embedding(num_classes, time_embedding_dim)

        make_res_block = lambda in_c, out_c, **kwargs: ResBlock(
            in_c,
            out_c,
            signal_dim=signal_dim,
            norm_eps=1.0e-5,
            time_embedding_channels=time_embedding_dim,
            dropout=dropout,
            use_checkpoint=use_checkpoint,
            use_scale_shift_norm=use_scale_shift_norm,
            **kwargs,
        )

        def make_attn_block(in_c: int) -> nn.Module:
            err_msg = "either `num_heads` or `num_head_channels` should be provided"
            if num_head_channels is not None:
                n_heads = in_c // num_head_channels
            else:
                if num_heads is None:
                    raise ValueError(err_msg)
                n_heads = num_heads
            head_c = in_c // n_heads if num_head_channels is None else num_head_channels
            if not use_spatial_transformer:
                # TODO: support attn_split_chunk
                return MultiHeadSpatialAttention(
                    in_c,
                    num_heads=n_heads,
                    num_head_channels=head_c,
                    use_checkpoint=use_checkpoint,
                )
            return SpatialTransformer(
                in_c,
                n_heads,
                head_c,
                num_layers=num_transformer_layers,
                context_dim=context_dim,
                use_linear=use_linear_in_transformer,
                use_checkpoint=use_checkpoint,
                attn_split_chunk=attn_split_chunk,
                tome_info=tome_info,
            )

        def make_downsample(in_c: int, out_c: int) -> TimestepAttnSequential:
            if not resample_with_resblock:
                return TimestepAttnSequential(
                    ResDownsample(
                        in_c,
                        resample_with_conv,
                        signal_dim=signal_dim,
                        out_channels=out_c,
                    )
                )
            res_block = make_res_block(in_c, out_c, integrate_downsample=True)
            return TimestepAttnSequential(res_block)

        # input
        input_blocks = [
            TimestepAttnSequential(
                conv_nd(signal_dim, in_channels, start_channels, 3, padding=1)
            ),
        ]
        zero_convs = [self.make_zero_conv(start_channels)]

        self.input_hint_block = TimestepAttnSequential(
            conv_nd(signal_dim, hint_channels, 16, 3, padding=1),
            nn.SiLU(),
            conv_nd(signal_dim, 16, 16, 3, padding=1),
            nn.SiLU(),
            conv_nd(signal_dim, 16, 32, 3, padding=1, stride=2),
            nn.SiLU(),
            conv_nd(signal_dim, 32, 32, 3, padding=1),
            nn.SiLU(),
            conv_nd(signal_dim, 32, 96, 3, padding=1, stride=2),
            nn.SiLU(),
            conv_nd(signal_dim, 96, 96, 3, padding=1),
            nn.SiLU(),
            conv_nd(signal_dim, 96, 256, 3, padding=1, stride=2),
            nn.SiLU(),
            zero_module(conv_nd(signal_dim, 256, start_channels, 3, padding=1)),
        )

        feature_size = start_channels
        input_block_channels = [start_channels]
        in_nc = start_channels
        downsample_rate = 1
        for i, multiplier in enumerate(channel_multipliers):
            for _ in range(num_res_blocks):
                out_nc = multiplier * start_channels
                blocks = [make_res_block(in_nc, out_nc)]
                in_nc = out_nc
                if downsample_rate in attention_downsample_rates:
                    blocks.append(make_attn_block(in_nc))
                input_blocks.append(TimestepAttnSequential(*blocks))
                zero_convs.append(self.make_zero_conv(in_nc))
                feature_size += in_nc
                input_block_channels.append(in_nc)
            if i != len(channel_multipliers) - 1:
                out_nc = in_nc
                input_blocks.append(make_downsample(in_nc, out_nc))
                zero_convs.append(self.make_zero_conv(in_nc))
                downsample_rate *= 2
                feature_size += in_nc
                input_block_channels.append(in_nc)

        self.input_blocks = nn.ModuleList(input_blocks)
        self.zero_convs = nn.ModuleList(zero_convs)

        # middle_block
        self.middle_block = TimestepAttnSequential(
            make_res_block(in_nc, in_nc),
            make_attn_block(in_nc),
            make_res_block(in_nc, in_nc),
        )
        self.middle_block_out = self.make_zero_conv(in_nc)
        feature_size += in_nc

    def make_zero_conv(self, channels: int) -> TimestepAttnSequential:
        return TimestepAttnSequential(
            zero_module(conv_nd(self.signal_dim, channels, channels, 1, padding=0))
        )

    def set_tome_info(self, tome_info: Optional[Dict[str, Any]]) -> None:
        for m in self.modules():
            if isinstance(m, SpatialTransformer):
                m.set_tome_info(tome_info)

    def forward(
        self,
        net: Tensor,
        hint: Tensor,
        *,
        timesteps: Tensor,
        context: Optional[Tensor] = None,
    ) -> List[Tensor]:
        # tomesd
        for m in self.modules():
            if isinstance(m, SpatialTransformer):
                for block in m.blocks:
                    if block.tome_info is not None:
                        block.tome_info["size"] = net.shape[-2:]

        # timenet
        time_net = timestep_embedding(
            timesteps,
            self.start_channels,
            dtype=net.dtype,
            repeat_only=False,
        )
        time_net = self.time_embed(time_net)
        guided_hint = self.input_hint_block(hint, time_net, context)

        # main
        nets = []
        for block, zero_conv in zip(self.input_blocks, self.zero_convs):
            net = block(net, time_net, context)
            if guided_hint is not None:
                net += guided_hint
                guided_hint = None
            nets.append(zero_conv(net, time_net, context))

        net = self.middle_block(net, time_net, context)
        net = self.middle_block_out(net, time_net, context)
        nets.append(net)

        return nets


__all__ = [
    "UNetDiffuser",
    "ControlNet",
]



import gc
import time
import torch

import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from typing import Generic
from typing import TypeVar
from typing import Callable
from typing import Optional
from cftool.misc import TIME_FORMAT
from cftool.misc import print_info
from cftool.misc import sort_dict_by_value
from cftool.types import tensor_dict_type
from torch.cuda.amp.autocast_mode import autocast

from .zoo import TPipeline
from ..misc.toolkit import is_cpu
from ..misc.toolkit import get_device
from ..misc.toolkit import empty_cuda_cache


TAPI = TypeVar("TAPI", bound="APIMixin")
TItem = TypeVar("TItem")


class APIMixin:
    m: nn.Module
    device: torch.device
    use_amp: bool
    use_half: bool

    def __init__(
        self,
        m: nn.Module,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ):
        if use_amp and use_half:
            raise ValueError("`use_amp` & `use_half` should not be True simultaneously")
        self.m = m.eval().requires_grad_(False)
        self.device = device
        self.use_amp = use_amp
        self.use_half = use_half

    def to(
        self,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> None:
        if use_amp and use_half:
            raise ValueError("`use_amp` & `use_half` should not be True simultaneously")
        self.device = device
        self.use_amp = use_amp
        self.use_half = use_half
        device_is_cpu = is_cpu(device)
        if device_is_cpu:
            self.m.to(device)
        if use_half:
            self.m.half()
        else:
            self.m.float()
        if not device_is_cpu:
            self.m.to(device)

    def empty_cuda_cache(self) -> None:
        empty_cuda_cache(self.device)

    @property
    def amp_context(self) -> autocast:
        return autocast(enabled=self.use_amp)

    @classmethod
    def from_pipeline(
        cls: Type[TAPI],
        m: TPipeline,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
        **kwargs: Any,
    ) -> TAPI:
        if use_amp and use_half:
            raise ValueError("`use_amp` & `use_half` should not be True simultaneously")
        model = m.build_model.model
        if use_half:
            model.half()
        if device is not None:
            model.to(device)
        return cls(
            model,
            get_device(model),
            use_amp=use_amp,
            use_half=use_half,
            **kwargs,
        )


class ILoadableItem(Generic[TItem]):
    _item: Optional[TItem]

    def __init__(self, init_fn: Callable[[], TItem], *, init: bool = False):
        self.init_fn = init_fn
        self.load_time = time.time()
        self._item = init_fn() if init else None

    def load(self, **kwargs: Any) -> TItem:
        self.load_time = time.time()
        if self._item is None:
            self._item = self.init_fn()
        return self._item

    def unload(self) -> None:
        self._item = None
        gc.collect()


class ILoadablePool(Generic[TItem], metaclass=ABCMeta):
    pool: Dict[str, ILoadableItem]
    activated: Dict[str, ILoadableItem]

    # set `limit` to negative values to indicate 'no limit'
    def __init__(self, limit: int = -1):
        self.pool = {}
        self.activated = {}
        self.limit = limit
        if limit == 0:
            raise ValueError(
                "limit should either be negative "
                "(which indicates 'no limit') or be positive"
            )

    def __contains__(self, key: str) -> bool:
        return key in self.pool

    def register(self, key: str, init_fn: Callable[[bool], ILoadableItem]) -> None:
        if key in self.pool:
            raise ValueError(f"key '{key}' already exists")
        init = self.limit < 0 or len(self.activated) < self.limit
        loadable_item = init_fn(init)
        self.pool[key] = loadable_item
        if init:
            self.activated[key] = loadable_item

    def get(self, key: str, **kwargs: Any) -> TItem:
        loadable_item = self.pool.get(key)
        if loadable_item is None:
            raise ValueError(f"key '{key}' does not exist")
        item = loadable_item.load(**kwargs)
        if key in self.activated:
            return item
        load_times = {k: v.load_time for k, v in self.activated.items()}
        earliest_key = list(sort_dict_by_value(load_times).keys())[0]
        self.activated.pop(earliest_key).unload()
        self.activated[key] = loadable_item

        time_format = "-".join(TIME_FORMAT.split("-")[:-1])
        print_info(
            f"'{earliest_key}' is unloaded to make room for '{key}' "
            f"(last updated: {time.strftime(time_format, time.localtime(loadable_item.load_time))})"
        )
        return item


class Weights(ILoadableItem[tensor_dict_type]):
    def __init__(self, path: str, *, init: bool = False):
        super().__init__(lambda: torch.load(path), init=init)


class WeightsPool(ILoadablePool[tensor_dict_type]):
    def register(self, key: str, path: str) -> None:  # type: ignore
        super().register(key, lambda init: Weights(path, init=init))



import numpy as np

from typing import Any
from cftool.misc import print_warning
from cftool.array import iou
from cftool.array import corr
from cftool.array import softmax
from cftool.array import get_full_logits
from cftool.array import get_label_predictions

from .schema import IMetric

try:
    from sklearn import metrics
except:
    metrics = None


@IMetric.register("acc")
class Accuracy(IMetric):
    def __init__(self, threshold: float = 0.5):
        super().__init__()
        self.threshold = threshold

    @property
    def is_positive(self) -> bool:
        return True

    def forward(self, logits: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        predictions = get_label_predictions(logits, self.threshold)
        return (predictions == labels).mean().item()


@IMetric.register("quantile")
class Quantile(IMetric):
    def __init__(self, q: Any):
        super().__init__()
        if not isinstance(q, float):
            q = np.asarray(q, np.float32).reshape([1, -1])
        self.q = q

    @property
    def is_positive(self) -> bool:
        return False

    def forward(self, predictions: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        diff = labels - predictions
        return np.maximum(self.q * diff, (self.q - 1.0) * diff).mean(0).sum().item()


@IMetric.register("f1")
class F1Score(IMetric):
    def __init__(self, average: str = "macro"):
        super().__init__()
        self.average = average
        if metrics is None:
            print_warning("`scikit-learn` needs to be installed for `F1Score`")

    @property
    def is_positive(self) -> bool:
        return True

    @property
    def requires_all(self) -> bool:
        return True

    def forward(self, logits: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        if metrics is None:
            return 0.0
        logits = get_full_logits(logits)
        num_classes = logits.shape[1]
        classes = logits.argmax(1)
        labels = labels.ravel()
        if num_classes == 2:
            return metrics.f1_score(labels, classes)
        return metrics.f1_score(labels, classes, average=self.average)


@IMetric.register("r2")
class R2Score(IMetric):
    def __init__(self) -> None:
        super().__init__()
        if metrics is None:
            print_warning("`scikit-learn` needs to be installed for `R2Score`")

    @property
    def is_positive(self) -> bool:
        return True

    def forward(self, predictions: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        if metrics is None:
            return 0.0
        return metrics.r2_score(labels, predictions)


@IMetric.register("auc")
class AUC(IMetric):
    def __init__(self) -> None:
        super().__init__()
        if metrics is None:
            print_warning("`scikit-learn` needs to be installed for `AUC`")

    @property
    def is_positive(self) -> bool:
        return True

    @property
    def requires_all(self) -> bool:
        return True

    def forward(self, logits: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        if metrics is None:
            return 0.0
        logits = get_full_logits(logits)
        num_classes = logits.shape[1]
        probabilities = softmax(logits)
        labels = labels.ravel()
        if num_classes == 2:
            return metrics.roc_auc_score(labels, probabilities[..., 1])
        return metrics.roc_auc_score(labels, probabilities, multi_class="ovr")


@IMetric.register("mae")
class MAE(IMetric):
    @property
    def is_positive(self) -> bool:
        return False

    def forward(self, predictions: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        return np.mean(np.abs(labels - predictions)).item()


@IMetric.register("mse")
class MSE(IMetric):
    @property
    def is_positive(self) -> bool:
        return False

    def forward(self, predictions: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        return np.mean(np.square(labels - predictions)).item()


@IMetric.register("ber")
class BER(IMetric):
    def __init__(self) -> None:
        super().__init__()
        if metrics is None:
            print_warning("`scikit-learn` needs to be installed for `AUC`")

    @property
    def is_positive(self) -> bool:
        return False

    def forward(self, predictions: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        if metrics is None:
            return 0.0
        mat = metrics.confusion_matrix(labels.ravel(), predictions.ravel())
        tp = np.diag(mat)
        fp = mat.sum(axis=0) - tp
        fn = mat.sum(axis=1) - tp
        tn = mat.sum() - (tp + fp + fn)
        return (0.5 * np.mean((fn / (tp + fn) + fp / (tn + fp)))).item()


@IMetric.register("corr")
class Correlation(IMetric):
    @property
    def is_positive(self) -> bool:
        return True

    def forward(self, predictions: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        return corr(predictions, labels, get_diagonal=True).mean().item()


@IMetric.register("iou")
class IOU(IMetric):
    @property
    def is_positive(self) -> bool:
        return True

    def forward(self, logits: np.ndarray, labels: np.ndarray) -> float:  # type: ignore
        return iou(logits, labels).mean().item()


__all__ = [
    "AUC",
    "BER",
    "IOU",
    "MAE",
    "MSE",
    "F1Score",
    "R2Score",
    "Accuracy",
    "Quantile",
    "Correlation",
]



import numpy as np

from typing import Any
from typing import Dict
from cftool.array import is_float

from .recognizer import RecognizerBlock
from ....schema import DataBundle
from ....schema import ColumnTypes
from ....schema import INoInitDataBlock


@INoInitDataBlock.register("ml_gather")
class GatherBlock(INoInitDataBlock):
    num_features: int
    num_labels: int

    def to_info(self) -> Dict[str, Any]:
        return dict(num_features=self.num_features, num_labels=self.num_labels)

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        return bundle

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        if not isinstance(bundle.x_train, np.ndarray):
            self.num_features = 0
        else:
            self.num_features = bundle.x_train.shape[-1]
        if bundle.y_train is None or not isinstance(bundle.y_train, np.ndarray):
            self.num_labels = 0
            return bundle
        y_dim = bundle.y_train.shape[-1]
        if y_dim > 1:
            self.num_labels = y_dim
        else:
            b_recognizer = self.try_get_previous(RecognizerBlock)
            if b_recognizer is None:
                if is_float(bundle.y_train):
                    self.num_labels = 1
                else:
                    self.num_labels = len(np.unique(bundle.y_train))
            else:
                key, ctype = list(b_recognizer.label_types.items())[0]
                if ctype == ColumnTypes.NUMERICAL:
                    self.num_labels = 1
                else:
                    self.num_labels = b_recognizer.num_unique_labels[key]
        return bundle


__all__ = ["GatherBlock"]



from typing import Any
from typing import Dict
from typing import Optional

from .schema import IVanillaGAN
from .discriminators import DiscriminatorBase
from ..decoder import make_decoder


@IVanillaGAN.register("gan")
class VanillaGAN(IVanillaGAN):
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        latent_dim: int = 64,
        latent_resolution: int = 7,
        *,
        generator: str = "vanilla",
        discriminator: str = "basic",
        generator_config: Optional[Dict[str, Any]] = None,
        discriminator_config: Optional[Dict[str, Any]] = None,
        num_classes: Optional[int] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.latent_dim = latent_dim
        self.build_loss(
            num_classes=num_classes,
            gan_mode=gan_mode,
            gan_loss_config=gan_loss_config,
        )
        # generator
        if generator_config is None:
            generator_config = {}
        generator_config["img_size"] = img_size
        generator_config["latent_dim"] = latent_dim
        generator_config["latent_resolution"] = latent_resolution
        generator_config["out_channels"] = out_channels or in_channels
        generator_config["num_classes"] = num_classes
        self.generator = make_decoder(generator, generator_config, is_1d=True)  # type: ignore
        # discriminator
        if discriminator_config is None:
            discriminator_config = {}
        discriminator_config["in_channels"] = in_channels
        discriminator_config["num_classes"] = num_classes
        self.discriminator = DiscriminatorBase.make(
            discriminator,
            config=discriminator_config,
        )


__all__ = ["VanillaGAN"]
